#
# rotate.py -- mode for rotating images
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
from ginga.ImageView import ImageViewBase
import math

from ginga import trcalc
from ginga.util import wcs
from ginga.modes.mode_base import Mode


class RotateMode(Mode):
    """
    Rotate Mode enables bindings that can flip or swap the axes of an
    image, or rotate it, in a Ginga image viewer.

    Enter the mode by
    -----------------
    * Space, then "r"

    Exit the mode by
    ----------------
    * Esc

    Default bindings in mode
    ------------------------
    * "[" : (toggle) flip the image in the X axis
    * "{" : Restore the X axis
    * "]" : (toggle) flip the image in the Y axis
    * "}" : Restore the Y axis
    * backslash : (toggle) swap the X and Y axes
    * "|" : Restore the swapped axes to normal
    * R : reset rotation to 0 deg (does not reset any flips or swaps)
    * t : resets any flips or swaps
    * period : rotate image incrementally by +90 deg
    * comma : rotate image by incrementally -90 deg
    * o : orient the image so that North points up and East points left
    * O : orient the image so that North points up and East points right
    * left drag : rotate the image interactively
    * right click : reset the rotation to 0 deg (same as R)
    * rotation gesture : rotate the image interactively

    """

    @classmethod
    def is_compatible_viewer(cls, viewer):
        return isinstance(viewer, ImageViewBase)

    def __init__(self, viewer, settings=None):
        super().__init__(viewer, settings=settings)

        self.actions = dict(
            dmod_rotate=['__r', None, None],

            kp_flip_x=['[', '{', 'rotate+[', 'rotate+{'],
            kp_flip_y=[']', '}', 'rotate+]', 'rotate+}'],
            kp_swap_xy=['backslash', '|', 'rotate+backslash', 'rotate+|'],
            kp_rotate_reset=['R', 'rotate+r', 'rotate+R'],
            kp_transform_reset=['rotate+t'],
            kp_rotate_inc90=['.', 'rotate+.'],
            kp_rotate_dec90=[',', 'rotate+,'],
            kp_orient_lh=['o', 'rotate+o'],
            kp_orient_rh=['O', 'rotate+O'],

            ms_rotate=['rotate+left'],
            ms_rotate_reset=['rotate+right'],
            pi_rotate=['rotate+pinch'],

            mouse_rotate_acceleration=0.75,
            pinch_rotate_acceleration=1.0)

        self._start_rot = 0

    def __str__(self):
        return 'rotate'

    @property
    def canflip(self):
        bd = self.viewer.get_bindings()
        return bd.get_feature_allow('flip')

    @property
    def canrotate(self):
        bd = self.viewer.get_bindings()
        return bd.get_feature_allow('rotate')

    def start(self):
        pass

    def stop(self):
        self.onscreen_message(None)

    def _rotate_xy(self, viewer, x, y, msg=True):
        msg = self.settings.get('msg_rotate', msg)
        ctr_x, ctr_y = viewer.get_center()
        if None in (self._start_x, self._start_y):
            # missed button down event, most likely, or we're getting this
            # motion callback too early
            return
        deg1 = math.degrees(math.atan2(ctr_y - self._start_y,
                                       self._start_x - ctr_x))
        deg2 = math.degrees(math.atan2(ctr_y - y, x - ctr_x))
        delta_deg = deg2 - deg1
        deg = math.fmod(self._start_rot + delta_deg, 360.0)
        if msg:
            self.onscreen_message("Rotate: %.2f" % (deg))
        viewer.rotate(deg)

    def _rotate_inc(self, viewer, inc_deg, msg=True):
        msg = self.settings.get('msg_rotate_inc', msg)
        cur_rot_deg = viewer.get_rotation()
        rot_deg = math.fmod(cur_rot_deg + inc_deg, 360.0)
        viewer.rotate(rot_deg)
        if msg:
            self.onscreen_message("Rotate Inc: (%.2f) %.2f" % (
                inc_deg, rot_deg), delay=1.0)

    def _orient(self, viewer, righthand=False, msg=True):
        msg = self.settings.get('msg_orient', msg)
        image = viewer.get_image()

        (x, y, xn, yn, xe, ye) = wcs.calc_compass_center(image)
        degn = math.degrees(math.atan2(xn - x, yn - y))
        self.logger.info("degn=%f xe=%f ye=%f" % (
            degn, xe, ye))
        # rotate east point also by degn
        xe2, ye2 = trcalc.rotate_pt(xe, ye, degn, xoff=x, yoff=y)
        dege = math.degrees(math.atan2(xe2 - x, ye2 - y))
        self.logger.info("dege=%f xe2=%f ye2=%f" % (
            dege, xe2, ye2))

        # if right-hand image, flip it to make left hand
        xflip = righthand
        if dege > 0.0:
            xflip = not xflip
        if xflip:
            degn = - degn

        viewer.transform(xflip, False, False)
        viewer.rotate(degn)
        if msg:
            self.onscreen_message("Orient: rot=%.2f flipx=%s" % (
                degn, str(xflip)), delay=1.0)

    #####  KEYBOARD ACTION CALLBACKS #####

    def kp_flip_x(self, viewer, event, data_x, data_y, msg=True):
        if not self.canflip:
            return False
        event.accept()
        msg = self.settings.get('msg_transform', msg)
        flipX, flipY, swapXY = viewer.get_transforms()
        if event.key == '[':
            flipx = not flipX
        else:
            flipx = False
        viewer.transform(flipx, flipY, swapXY)
        if msg:
            self.onscreen_message("Flip X=%s" % flipx, delay=1.0)

    def kp_flip_y(self, viewer, event, data_x, data_y, msg=True):
        if not self.canflip:
            return False
        event.accept()
        msg = self.settings.get('msg_transform', msg)
        flipX, flipY, swapXY = viewer.get_transforms()
        if event.key == ']':
            flipy = not flipY
        else:
            flipy = False
        viewer.transform(flipX, flipy, swapXY)
        if msg:
            self.onscreen_message("Flip Y=%s" % flipy, delay=1.0)

    def kp_swap_xy(self, viewer, event, data_x, data_y, msg=True):
        if not self.canflip:
            return False
        event.accept()
        msg = self.settings.get('msg_transform', msg)
        flipX, flipY, swapXY = viewer.get_transforms()
        if event.key == 'backslash':
            swapxy = not swapXY
        else:
            swapxy = False
        viewer.transform(flipX, flipY, swapxy)
        if msg:
            self.onscreen_message("Swap XY=%s" % swapxy, delay=1.0)

    def kp_transform_reset(self, viewer, event, data_x, data_y):
        if not self.canflip:
            return False
        event.accept()
        viewer.transform(False, False, False)
        self.onscreen_message("Flips/swaps reset", delay=0.5)

    def kp_rotate_reset(self, viewer, event, data_x, data_y):
        if not self.canrotate:
            return False
        event.accept()
        viewer.rotate(0.0)

    def kp_rotate_inc90(self, viewer, event, data_x, data_y, msg=True):
        if not self.canrotate:
            return False
        event.accept()
        self._rotate_inc(viewer, 90.0, msg=msg)

    def kp_rotate_dec90(self, viewer, event, data_x, data_y, msg=True):
        if not self.canrotate:
            return False
        event.accept()
        self._rotate_inc(viewer, -90.0, msg=msg)

    def kp_orient_lh(self, viewer, event, data_x, data_y, msg=True):
        if not self.canrotate:
            return False
        event.accept()
        self._orient(viewer, righthand=False, msg=msg)

    def kp_orient_rh(self, viewer, event, data_x, data_y, msg=True):
        if not self.canrotate:
            return False
        event.accept()
        self._orient(viewer, righthand=True, msg=msg)

    #####  SCROLL ACTION CALLBACKS #####

    #####  MOUSE ACTION CALLBACKS #####

    def ms_rotate(self, viewer, event, data_x, data_y, msg=True):
        """Rotate the image by dragging the cursor left or right.
        """
        if not self.canrotate:
            return False
        event.accept()
        msg = self.settings.get('msg_rotate', msg)

        x, y = self.get_win_xy(viewer)

        if event.state == 'move':
            self._rotate_xy(viewer, x, y)

        elif event.state == 'down':
            if msg:
                self.onscreen_message("Rotate (drag around center)",
                                      delay=1.0)
            self._start_x, self._start_y = x, y
            self._start_rot = viewer.get_rotation()

        else:
            self.onscreen_message(None)

    def ms_rotate_reset(self, viewer, event, data_x, data_y, msg=True):
        if not self.canrotate:
            return False
        event.accept()
        msg = self.settings.get('msg_rotate', msg)

        if event.state == 'down':
            viewer.rotate(0.0)
            self.onscreen_message("Rotation reset", delay=0.5)

    ##### GESTURE ACTION CALLBACKS #####

    # NOTE: rotation is currently handled by default in the pinch
    # gesture handler in the "pan" mode.  This function is here
    # in case one wanted to make a binding to only rotate and not
    # zoom as well via pinch.  There is no default binding for it.
    #
    def pi_rotate(self, viewer, event, msg=True):
        if not self.canrotate:
            return False
        event.accept()
        if event.state == 'start':
            self._start_rot = viewer.get_rotation()
        else:
            msg_str = None
            deg = self._start_rot - event.rot_deg
            rotate_accel = self.settings.get('pinch_rotate_acceleration', 1.0)
            deg = rotate_accel * deg
            viewer.rotate(deg)
            if msg_str is None:
                msg_str = "Rotate: %.2f" % (deg)
                msg = self.settings.get('msg_rotate', msg)

            if msg and msg_str is not None:
                self.onscreen_message(msg_str, delay=0.4)



#
# pipeline.py -- Base classes for pipelines in Ginga
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import time

from ginga.misc import Bunch, Callback
from ginga.util import action

__all__ = ['Pipeline']


class PipeError(Exception):
    pass


class Pipeline(Callback.Callbacks):

    def __init__(self, logger, stages, name=None):
        super(Pipeline, self).__init__()
        self.logger = logger
        self.cur_stage = None
        self._i = 0
        self.bboard = Bunch.Bunch()
        if name is None:
            name = 'noname'
        self.name = name
        self.enabled = True
        self.pipeline = list(stages)
        # undo/redo stack
        self.actions = action.ActionStack()

        for stage in self.pipeline:
            self._init_stage(stage)

        for name in ['pipeline-start',
                     'stage-executing', 'stage-errored', 'stage-done']:
            self.enable_callback(name)

    def _init_stage(self, stage):
        stage.pipeline = self
        stage.logger = self.logger
        stage.result = Bunch.Bunch(res_np=None)

    def insert(self, i, stage):
        self.pipeline.insert(i, stage)
        self._init_stage(stage)

    def append(self, stage):
        self.pipeline.append(stage)
        self._init_stage(stage)

    def remove(self, stage):
        stage.pipeline = None
        self.pipeline.remove(stage)

    def enable(self, tf):
        self.enabled = tf

    def run_stage_idx(self, i):
        if not self.enabled:
            self.logger.info("pipeline disabled")

        if i < 0 or i >= len(self.pipeline):
            raise ValueError("No stage at index {}".format(i))

        stage = self.pipeline[i]
        prev_stage = None if i == 0 else self.pipeline[i - 1]
        self.cur_stage = stage
        self.make_callback('stage-executing', stage)
        start_time = time.time()
        try:
            stage.run(prev_stage)
            stop_time = time.time()
            self.make_callback('stage-done', stage)

        except Exception as e:
            stop_time = time.time()
            self.logger.error("Error running stage %d (%s): %s" % (
                i, str(stage), e), exc_info=True)
            self.stop()
            self.make_callback('stage-errored', stage)

        self.logger.debug("stage '%s' took %.4f sec" % (stage._stagename,
                                                        stop_time - start_time))

    def run_from(self, stage):
        self.make_callback('pipeline-start', stage)
        self._i = self.pipeline.index(stage)
        start_time = time.time()

        while self._i < len(self.pipeline):
            self.run_stage_idx(self._i)
            self._i += 1

        stop_time = time.time()
        self.logger.debug("pipeline '%s' total execution %.4f sec" % (
            self.name, stop_time - start_time))

    def run_all(self):
        self.run_from(self.pipeline[0])

    def stop(self):
        self._i = len(self.pipeline)

    def get_data(self, stage):
        return stage.result.res_np

    def send(self, **kwargs):
        self.cur_stage.result = Bunch.Bunch(kwargs)

    def set(self, **kwargs):
        self.bboard.setvals(**kwargs)

    def get(self, *args):
        if len(args) == 1:
            return self.bboard[args[0]]
        if len(args) == 2:
            return self.bboard.get(args[0], args[1])
        raise ValueError("Pass keyword as argument")

    def invalidate(self):
        for stage in self.pipeline:
            stage.invalidate()

    def push(self, act):
        self.actions.push(act)

    def undo(self):
        act = self.actions.undo()
        self.run_from(act.obj)

    def redo(self):
        act = self.actions.redo()
        self.run_from(act.obj)

    def save(self):
        dd = dict(name=self.name, enabled=self.enabled)
        d = dict(pipeline=dd)
        l = [stage.export_as_dict() for stage in self.pipeline]
        dd['stages'] = l
        return d

    def load(self, d, cd):
        # reinitialize some things
        self.cur_stage = None
        self._i = 0
        self.bboard = Bunch.Bunch()
        self.actions = action.ActionStack()

        dd = d['pipeline']
        self.name = dd['name']
        self.enabled = dd.get('enabled', True)
        l = []
        for sd in dd['stages']:
            # instantiate stage
            stage = cd[sd['type']]()
            self._init_stage(stage)
            stage.import_from_dict(sd)
            l.append(stage)
        self.pipeline = l

        return d

    def __getitem__(self, idx):
        return self.pipeline[idx]

    def __contains__(self, stage):
        return stage in self.pipeline

    def index(self, stage):
        return self.pipeline.index(stage)

    def __len__(self):
        return len(self.pipeline)



#
# CvHelp.py -- help classes for the Cv drawing
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.

import numpy as np
import cv2

from ginga.fonts import font_asst


def get_cached_font(fontname, fontsize):
    key = ('opencv', fontname, fontsize)
    try:
        return font_asst.get_cache(key)

    except KeyError:
        # see if we can build the font
        info = font_asst.get_font_info(fontname, subst_ok=True)

        font = cv2.freetype.createFreeType2()
        font.loadFontData(info.font_path, id=0)
        font_asst.add_cache(key, font)

        return font


def load_font(font_name, font_file):
    if not font_asst.have_font(font_name):
        font_asst.add_font(font_file, font_name=font_name)
    return font_name


class CvContext(object):

    def __init__(self, canvas):
        self.canvas = canvas

    def set_canvas(self, canvas):
        self.canvas = canvas

    def text_extents(self, text, font):
        _font, _scale = font.render.font, font.render.scale
        retval, baseline = _font.getTextSize(text, _scale, -1)
        wd_px, ht_px = retval
        return wd_px, ht_px

    def image(self, pt, rgb_arr):
        # TODO: is there a faster way to copy this array in?
        cx, cy = pt[:2]
        daht, dawd, depth = rgb_arr.shape

        self.canvas[cy:cy + daht, cx:cx + dawd, :] = rgb_arr

    def text(self, pt, text, font, line, fill):
        x, y = pt
        if font is not None and fill is not None:
            _font, _scale = font.render.font, font.render.scale
            _font.putText(self.canvas, text, (x, y), _scale,
                          # text is not filled unless thickness is negative
                          fill.render.color, thickness=-1,
                          line_type=cv2.LINE_AA, bottomLeftOrigin=True)

    def line(self, pt1, pt2, line):
        if line is not None:
            x1, y1 = int(round(pt1[0])), int(round(pt1[1]))
            x2, y2 = int(round(pt2[0])), int(round(pt2[1]))
            cv2.line(self.canvas, (x1, y1), (x2, y2), line.render.color,
                     line.linewidth)

    def circle(self, pt, radius, line, fill):
        x, y = pt
        radius = int(radius)
        if fill is not None:
            cv2.circle(self.canvas, (x, y), radius, fill.render.color, -1)
        if line is not None:
            cv2.circle(self.canvas, (x, y), radius, line.render.color,
                       line.linewidth)

    def polygon(self, points, line, fill):
        pts = np.array(points, np.int32)
        pts = pts.reshape((-1, 1, 2))
        if fill is not None:
            cv2.fillPoly(self.canvas, [pts], fill.render.color)
        if line is not None:
            cv2.polylines(self.canvas, [pts], True, line.render.color,
                          line.linewidth)

    def path(self, points, line):
        pts = np.array(points, np.int32)
        pts = pts.reshape((-1, 1, 2))
        if line is not None:
            cv2.polylines(self.canvas, [pts], False, line.render.color,
                          line.linewidth)


# END



#! /usr/bin/env python3
#
"""
Plot time series data in an interactive plot viewer.

Usage
=====
$ python3 plot_time_series.py --loglevel=20 --stderr

Plots add new data every second, and update on screen every 5 sec.
24 hours of data is kept.  Each plot starts out in "autoaxis X PAN"
and "autoaxis Y VIS".

Things you can do in a plot:

1) Scroll the mouse wheel to zoom the X axis

2) Hold CTRL and scroll the mouse wheel to zoom the Y axis.

3) Press 'y' to autozoom the Y axis to the full range of the Y data.
   (The message "Autoaxis Y ON" should appear briefly).
   Press 'y' again to toggle off this behavior.

4) Press 'v' to autozoom the Y axis to the range of the *visible* data shown
   (The message "Autoaxis Y VIS" should appear briefly).
   Press 'v' again to toggle off this behavior.

5) Press 'x' to autozoom the X axis to the full range of the X data
   (The message "Autoaxis X ON" should appear briefly).
   Press 'x' again to toggle off this behavior.

6) Press 'p' to autopan the X axis to always show the latest data on
   the right (the message "Autoaxis X PAN" should appear briefly).
   Press 'p' again to toggle off this behavior.
.
"""
import sys
import time
import threading

import numpy as np

import ginga.toolkit as ginga_toolkit
from ginga.misc import log, Bunch
from ginga.plot.plotaide import PlotAide
import ginga.plot.data_source as dsp

win_wd, win_ht = 800, 280


class FakeData:
    """Generate fake time-series data."""

    def __init__(self, name, t_start, y_range, num_pts):
        self.name = name
        self.t_start = t_start
        self.y_range = y_range
        self.num_pts = num_pts
        self.data_src = None
        self.tv = 0.0
        self.tv_dir = 'up'
        self.tv_dct = {0: 'up', 1: 'down'}
        self.tv_delta = 0.1
        self.tv_dmin = 1.0
        self.tv_dmax = 30.0
        self.tv_deadline = time.time()

    def rand_rng(self, a, b):
        return (b - a) * np.random.random_sample() + a

    def generate_point(self, t):
        x, y = t, self.tv

        if self.tv_dir == 'up':
            y = self.rand_rng(y, min(y + self.tv_delta, self.y_range[1]))
        else:
            y = self.rand_rng(max(y - self.tv_delta, self.y_range[0]), y)
        self.tv = y
        if t >= self.tv_deadline:
            v = np.random.randint(0, 2)
            self.tv_dir = self.tv_dct[v]
            self.tv_deadline = t + self.rand_rng(self.tv_dmin, self.tv_dmax)

        ## p = np.random.randint(0, 100)
        ## if p >= 98:
        ##     y = np.nan
        return (x, y)

    def init_points(self, data_src, start=None):
        N, t = self.num_pts, self.t_start
        if start is None:
            start = (self.y_range[0] + self.y_range[1]) * 0.5
        self.tv = start
        self.tv_deadline = t - N + self.rand_rng(self.tv_dmin, self.tv_dmax)
        points = np.array([self.generate_point(ti)
                           for ti in np.arange(t - N, t, 1.0)])
        data_src.set_points(points)
        self.data_src = data_src

    def add_point(self, t):
        pt = self.generate_point(t)
        self.data_src.add(pt)
        return pt


def timer1_cb(timer, fdg_l, interval):
    t = time.time()
    timer.set(interval)

    for fdg in fdg_l:
        fdg.add_point(t)
        dsp.update_plot_from_source(fdg.data_src, fdg.data_src.plot,
                                    update_limits=True)


def timer2_cb(timer, app, aides, fdg_l, interval):
    timer.set(interval)

    for a in aides:
        # keep plots responsive
        app.process_events()
        a.aide.update_plots()


def make_plot(logger, dims, sources, y_rng, y_acc=np.mean,
              title='', warn_y=None, alert_y=None,
              show_x_axis=True, show_y_axis=True):

    from ginga.gw import Viewers
    from ginga.canvas.types import plots as gplots
    import ginga.plot.time_series as tsp

    win_wd, win_ht = dims[:2]
    viewer = Viewers.CanvasView(logger, render='widget')

    viewer.set_desired_size(win_wd, win_ht)
    viewer.set_zoom_algorithm('rate')
    viewer.set_zoomrate(1.41)
    viewer.enable_autozoom('off')
    viewer.set_background('white')
    viewer.set_foreground('black')
    viewer.set_enter_focus(True)

    # our plot
    aide = PlotAide(viewer)
    aide.settings.set(autoaxis_x='pan', autoaxis_y='vis')

    bg = tsp.TimePlotBG(warn_y=warn_y, alert_y=alert_y, linewidth=2)
    aide.add_plot_decor(bg)

    title = tsp.TimePlotTitle(title=title)
    aide.add_plot_decor(title)

    x_axis = tsp.XTimeAxis(num_labels=4)
    aide.add_plot_decor(x_axis)

    y_axis = gplots.YAxis(num_labels=4)
    aide.add_plot_decor(y_axis)

    colors = ['purple', 'palegreen4', 'red', 'brown', 'blue']

    for i, src in enumerate(sources):
        psrc = gplots.XYPlot(name=src.name, color=colors[i % len(colors)],
                             x_acc=np.mean, y_acc=y_acc,
                             linewidth=2.0, coord='data')
        buf = np.zeros((src.num_pts, 2), dtype=float)
        dsrc = dsp.XYDataSource(buf, none_for_empty=True, overwrite=True)
        dsrc.plot = psrc
        src.init_points(dsrc)

        aide.add_plot(psrc)
        dsp.update_plot_from_source(dsrc, psrc, update_limits=True)

    # initially, show last 4 hours worth of data.
    t, _ = dsrc.get_latest()
    aide.zoom_limit_x(t - 4 * 3600, t)

    # add scrollbar interface around this viewer
    si = Viewers.GingaScrolledViewerWidget(viewer=viewer, width=win_wd,
                                           height=win_ht)
    aide.configure_scrollbars(si)

    res = Bunch.Bunch(viewer=viewer, aide=aide, widget=si)
    return res


def make_data(t, N, names, y_range):
    srcs = []
    for i, name in enumerate(names):
        fdg = FakeData(name, t, y_range, N)
        srcs.append(fdg)

    return srcs


def cross_connect_plots(plot_info):
    # cross connect the plots so that zooming or panning in X in one
    # does the same to all the others
    plot_info = list(plot_info)
    m_settings = plot_info[0].aide.settings
    for res_a in plot_info:
        for res_b in plot_info:
            if res_b != res_a:
                res_a.aide.add_callback('plot-zoom-x',
                                        res_b.aide.plot_zoom_x_cb)

        if res_a.aide.settings is not m_settings:
            m_settings.share_settings(res_a.aide.settings,
                                      keylist=['autoaxis_x'])


def main(options, args):

    logger = log.get_logger("example1", options=options)

    if options.toolkit is None:
        logger.error("Please choose a GUI toolkit with -t option")

    # decide our toolkit, then import
    ginga_toolkit.use(options.toolkit)

    # now we can import
    from ginga.gw import Widgets

    def quit(self, *args):
        logger.info("Top window closed.")
        sys.exit()

    ev_quit = threading.Event()

    app = Widgets.Application(logger=logger)
    app.add_callback('shutdown', quit)

    w = app.make_window("EnvMon")
    w.add_callback('close', quit)

    vbox = Widgets.VBox()
    vbox.set_spacing(1)

    dims = (win_wd, win_ht)

    # default: data every second for 24 hours
    N = options.numvalues
    M = options.numplots
    t = time.time()
    plots = []
    fdgs = []

    # make a plot of outside and dome wind speed
    y_rng = (0.0, 50.0)
    srcs = make_data(t, N, ["Outside", "Dome"], y_rng)
    fdgs.extend(srcs)
    res = make_plot(logger, dims, srcs, y_rng,
                    y_acc=np.mean, title="Wind Speed (m/s)")
    vbox.add_widget(res.widget, stretch=1)
    plots.append(res)

    # make a plot of outside and dome temperature
    y_rng = (-30.0, 50.0)
    srcs = make_data(t, N, ["Outside", "Dome"], y_rng)
    fdgs.extend(srcs)
    res = make_plot(logger, dims, srcs, y_rng,
                    y_acc=np.mean, title="Temperature (C)")
    vbox.add_widget(res.widget, stretch=1)
    plots.append(res)

    # make a plot of outside and dome humidity
    y_rng = (0.0, 100.0)
    srcs = make_data(t, N, ["Outside", "Dome"], y_rng)
    fdgs.extend(srcs)
    res = make_plot(logger, dims, srcs, y_rng,
                    y_acc=np.mean, title="Humidity (%)",
                    warn_y=70, alert_y=80)
    vbox.add_widget(res.widget, stretch=1)
    plots.append(res)

    # make a plot of outside and dome dew point
    y_rng = (-30.0, 50.0)
    srcs = make_data(t, N, ["Outside", "Dome"], y_rng)
    fdgs.extend(srcs)
    res = make_plot(logger, dims, srcs, y_rng,
                    y_acc=np.mean, title="M1 & Dew (C)")
    vbox.add_widget(res.widget, stretch=1)
    plots.append(res)

    # make a plot of front and rear top-ring wind speed
    y_rng = (0.0, 50.0)
    srcs = make_data(t, N, ["Front", "Rear"], y_rng)
    fdgs.extend(srcs)
    res = make_plot(logger, dims, srcs, y_rng,
                    y_acc=np.mean, title="Top Ring Wind (m/s)")
    vbox.add_widget(res.widget, stretch=1)
    plots.append(res)

    # cross connect plots so zooming/panning in X affects all plots
    cross_connect_plots(plots)

    hbox = Widgets.HBox()
    hbox.set_margins(4, 2, 4, 2)

    wquit = Widgets.Button("Quit")
    wquit.add_callback('activated', quit)

    hbox.add_widget(Widgets.Label(''), stretch=1)
    hbox.add_widget(wquit)
    vbox.add_widget(hbox, stretch=0)

    w.set_widget(vbox)

    # timer to add a point every second
    t1 = app.make_timer()
    t1.add_callback('expired', timer1_cb, fdgs, 1.0)
    t1.set(1.0)

    # timer to update the plot every interval seconds
    t2 = app.make_timer()
    t2.add_callback('expired', timer2_cb, app, plots, fdgs,
                    options.update_interval)
    t2.set(options.update_interval)

    w.resize(win_wd, win_ht * len(plots) + 50)
    w.show()

    app.mainloop()


if __name__ == "__main__":

    # Parse command line options
    from argparse import ArgumentParser

    argprs = ArgumentParser("test ginga plot")

    argprs.add_argument("--debug", dest="debug", default=False,
                        action="store_true",
                        help="Enter the pdb debugger on main()")
    argprs.add_argument("-n", "--numvalues", dest="numvalues", default=86400,
                        type=int,
                        help="Number of items to show per plot")
    argprs.add_argument("-m", "--numplots", dest="numplots", default=2,
                        type=int,
                        help="Number of plots to show per graph")
    argprs.add_argument("--profile", dest="profile", action="store_true",
                        default=False,
                        help="Run the profiler on main()")
    argprs.add_argument("-t", "--toolkit", dest="toolkit", metavar="NAME",
                        default='qt',
                        help="Choose GUI toolkit (gtk|qt)")
    argprs.add_argument("--update", dest="update_interval", default=5.0,
                        type=float,
                        help="Number of seconds between plot updates")
    log.addlogopts(argprs)

    (options, args) = argprs.parse_known_args(sys.argv[1:])

    main(options, args)



#
# CompoundMixin.py -- enable compound capabilities.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
from functools import reduce

import numpy as np

__all__ = ['CompoundMixin']


class CompoundMixin(object):
    """A CompoundMixin is a mixin class that makes an object that is an
    aggregation of other objects.

    It is used to make generic compound drawing types as well as (for example)
    layers of canvases on top of an image.
    """

    def __init__(self):
        # holds a list of objects to be drawn
        self.objects = []
        if not hasattr(self, 'crdmap'):
            self.crdmap = None
        if not hasattr(self, 'coord'):
            self.coord = None
        self.opaque = False
        self._contains_reduce = np.logical_or

    def __contains__(self, key):
        return key in self.objects

    def __len__(self):
        return len(self.objects)

    def get_llur(self):
        """
        Get lower-left and upper-right coordinates of the bounding box
        of this compound object.

        Returns
        -------
        x1, y1, x2, y2: a 4-tuple of the lower-left and upper-right coords
        """
        points = np.array([obj.get_llur() for obj in self.objects])
        t_ = points.T
        x1, y1 = t_[0].min(), t_[1].min()
        x2, y2 = t_[2].max(), t_[3].max()
        return (x1, y1, x2, y2)

    def contains_pts(self, pts):
        if len(pts) == 0:
            return np.array([], dtype=bool)
        if len(self.objects) == 0:
            x_arr, y_arr = np.asarray(pts).T
            return np.full(x_arr.shape, False, dtype=bool)

        return reduce(self._contains_reduce,
                      map(lambda obj: obj.contains_pts(pts), self.objects))

    def get_items_at(self, pt):
        res = []
        for obj in self.objects:
            if obj.is_compound() and not obj.opaque:
                # non-opaque compound object, list up compatible members
                res.extend(obj.get_items_at(pt))
            elif obj.contains_pt(pt):
                #res.insert(0, obj)
                res.append(obj)
        return res

    def get_objects_by_kind(self, kind):
        return filter(lambda obj: obj.kind == kind, self.objects)

    def get_objects_by_kinds(self, kinds):
        return filter(lambda obj: obj.kind in kinds, self.objects)

    def select_contains_pt(self, viewer, pt):
        for obj in self.objects:
            if obj.select_contains_pt(viewer, pt):
                return True
        return False

    def select_items_at(self, viewer, pt, test=None):
        res = []
        try:
            for obj in self.objects:
                if obj.is_compound() and not obj.opaque:
                    # non-opaque compound object, list up compatible members
                    res.extend(obj.select_items_at(viewer, pt, test=test))

                is_inside = obj.select_contains_pt(viewer, pt)
                if test is None:
                    if is_inside:
                        res.append(obj)
                elif test(obj, pt, is_inside):
                    # custom test
                    res.append(obj)
        except Exception as e:
            self.logger.error("error selecting object(s): {}".format(e),
                              exc_info=True)
            res = []
        return res

    def initialize(self, canvas, viewer, logger):
        super().initialize(canvas, viewer, logger)

        # initialize children
        for obj in self.objects:
            obj.initialize(self, viewer, logger)

    def inherit_from(self, obj):
        self.crdmap = obj.crdmap
        self.logger = obj.logger
        self.viewer = obj.viewer

    def is_compound(self):
        return True

    def use_coordmap(self, mapobj):
        super().use_coordmap(mapobj)

        for obj in self.objects:
            obj.use_coordmap(mapobj)

    def draw(self, viewer):
        for obj in self.objects:
            obj.draw(viewer)

    def get_objects(self):
        return self.objects

    def has_object(self, obj):
        return obj in self.objects

    def copy(self, share=[]):
        obj = super().copy(share=share)
        obj.objects = [obj.copy(share=share) for obj in self.objects]
        return obj

    def delete_object(self, obj):
        self.objects.remove(obj)

    def delete_objects(self, objects):
        for obj in objects:
            self.delete_object(obj)

    def delete_all_objects(self):
        self.objects.clear()

    def roll_objects(self, n):
        num = len(self.objects)
        if num == 0:
            return
        n = n % num
        self.objects = self.objects[-n:] + self.objects[:-n]

    def swap_objects(self):
        num = len(self.objects)
        if num >= 2:
            l = self.objects
            self.objects = l[:num - 2] + [l[num - 1], l[num - 2]]

    def set_attr_all(self, **kwdargs):
        for obj in self.objects:
            for attrname, val in kwdargs.items():
                if hasattr(obj, attrname):
                    setattr(obj, attrname, val)

    def add_object(self, obj, belowThis=None):

        if obj in self.objects:
            raise ValueError("object is already in this compound object")

        obj.initialize(self, self.viewer, self.logger)

        if belowThis is None:
            self.objects.append(obj)
        else:
            index = self.objects.index(belowThis)
            self.objects.insert(index, obj)

    def raise_object(self, obj, aboveThis=None):
        if aboveThis is None:
            # no reference object--move to top
            self.objects.remove(obj)
            self.objects.append(obj)
        else:
            # Force an error if the reference object doesn't exist in list
            index = self.objects.index(aboveThis)
            self.objects.remove(obj)
            index = self.objects.index(aboveThis)
            self.objects.insert(index + 1, obj)

    def lower_object(self, obj, belowThis=None):
        if belowThis is None:
            # no reference object--move to bottom
            self.objects.remove(obj)
            self.objects.insert(0, obj)
        else:
            # Force an error if the reference object doesn't exist in list
            index = self.objects.index(belowThis)
            self.objects.remove(obj)
            index = self.objects.index(belowThis)
            self.objects.insert(index, obj)

    def rotate_deg(self, thetas, offset):
        for obj in self.objects:
            obj.rotate_deg(thetas, offset)

    def move_delta_pt(self, off_pt):
        for obj in self.objects:
            obj.move_delta_pt(off_pt)

    def scale_by_factors(self, factors):
        for obj in self.objects:
            obj.scale_by_factors(factors)

    def get_reference_pt(self):
        # Reference point for a compound object is the average of all
        # it's contituents reference points
        points = np.asarray([obj.get_reference_pt()
                             for obj in self.objects])
        t_ = points.T
        x, y = np.average(t_[0]), np.average(t_[1])
        return (x, y)

    get_center_pt = get_reference_pt

    def get_points(self):
        res = []
        for obj in self.objects:
            res.extend(list(obj.get_points()))
        return res

    def setup_edit(self, detail):
        detail.center_pos = self.get_center_pt()

    def get_edit_points(self, viewer):
        move_pt, scale_pt, rotate_pt = self.get_move_scale_rotate_pts(viewer)
        # currently only move_pt is supported for compound objects
        return [move_pt]

    def set_edit_point(self, i, pt, detail):
        if i == 0:
            # move control point
            ctr_x, ctr_y = self.get_reference_pt()
            delta_pt = (pt[0] - ctr_x, pt[1] - ctr_y)
            self.move_delta_pt(delta_pt)
        else:
            raise ValueError("No point corresponding to index %d" % (i))



#
# EventMixin.py -- mixin class for handling events in matplotlib.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
from matplotlib.backend_tools import Cursors

from ginga.cursors import cursor_info
from ginga.mplw import MplHelp


class PlotEventMixin:
    """Mixin class for adding Matplotlib non-web toolkit UI events to
    a plot viewer.
    """

    def __init__(self):
        # for matplotlib key handling
        self._keytbl = {
            'shift': 'shift_l',
            'control': 'control_l',
            'alt': 'alt_l',
            'win': 'meta_l',  # windows key
            'cmd': 'meta_l',  # Command key on Macs
            '`': 'backquote',
            '"': 'doublequote',
            "'": 'singlequote',
            '\\': 'backslash',
            ' ': 'space',
            'enter': 'return',
            'pageup': 'page_up',
            'pagedown': 'page_down',
        }

        # For callbacks
        for name in ['motion', 'button-press', 'button-release',
                     'key-press', 'key-release', 'drag-drop',
                     'scroll', 'map', 'focus', 'enter', 'leave',
                     'pinch', 'pan',  # 'swipe', 'tap',
                     'pixel-info']:
            self.enable_callback(name)

        self.add_callback('pixel-info', self.pixel_info_cb)

    def connect_ui(self, mpl_canvas):
        connect = mpl_canvas.mpl_connect
        connect("motion_notify_event", self.motion_notify_event)
        connect("button_press_event", self.button_press_event)
        connect("button_release_event", self.button_release_event)
        connect("scroll_event", self.scroll_event)
        mpl_canvas.capture_scroll = True
        connect("figure_enter_event", self.enter_notify_event)
        connect("figure_leave_event", self.leave_notify_event)
        connect("key_press_event", self.key_down_event)
        connect("key_release_event", self.key_up_event)
        connect("resize_event", self.resize_event)

        self.timer_resize = MplHelp.Timer(mplcanvas=mpl_canvas)
        self.timer_resize.add_callback('expired',
                                       lambda t: self.delayed_resize())

        self.timer_msg = MplHelp.Timer(mplcanvas=mpl_canvas)
        self.timer_msg.add_callback('expired',
                                    lambda t: self.onscreen_message_off())

        # Define cursors
        cursor_names = cursor_info.get_cursor_names()
        # TODO: handle other cursor types
        cross = Cursors.POINTER
        for curname in cursor_names:
            curinfo = cursor_info.get_cursor_info(curname)
            self.define_cursor(curinfo.name, cross)
        self.switch_cursor('pick')

    def _get_modifiers(self, event):
        return event.modifiers

    def _get_button(self, event):
        try:
            btn = 0x1 << (event.button.value - 1)
        except Exception:
            btn = 0
        return btn

    def transkey(self, keyname):
        self.logger.debug("matplotlib keyname='%s'" % (keyname))
        if keyname is None:
            return keyname
        key = keyname
        if 'shift+' in key:
            key = key.replace('shift+', '')
        if 'ctrl+' in key:
            key = key.replace('ctrl+', '')
        if 'alt+' in key:
            key = key.replace('alt+', '')
        if 'meta+' in key:
            key = key.replace('meta+', '')
        key = self._keytbl.get(key, key)
        return key

    def get_key_table(self):
        return self._keytbl

    def _get_key(self, event):
        keyval = self.transkey(event.key)
        self.logger.debug("key event, mpl={}, key={}".format(event.key,
                                                             keyval))
        return keyval

    def scroll_event(self, event):
        button = self._get_button(event)
        if event.button == 'up':
            direction = 0.0
        elif event.button == 'down':
            direction = 180.0
        amount = event.step
        modifiers = self._get_modifiers(event)

        data_x, data_y = event.xdata, event.ydata
        self.set_last_data_xy(data_x, data_y)

        num_degrees = amount  # ???
        self.make_ui_callback_viewer(self, 'scroll',
                                     direction, num_degrees,
                                     data_x, data_y)

    def button_press_event(self, event):
        button = self._get_button(event)
        modifiers = self._get_modifiers(event)
        # NOTE: event.xdata, event.ydata seem to be None
        last_data_x, last_data_y = self.get_last_data_xy()

        self.make_ui_callback_viewer(self, 'button-press', button,
                                     last_data_x, last_data_y)

    def button_release_event(self, event):
        button = self._get_button(event)
        modifiers = self._get_modifiers(event)
        # NOTE: event.xdata, event.ydata seem to be None
        last_data_x, last_data_y = self.get_last_data_xy()

        self.make_ui_callback_viewer(self, 'button-release', button,
                                     last_data_x, last_data_y)

    def motion_notify_event(self, event):
        button = self._get_button(event)
        modifiers = self._get_modifiers(event)

        data_x, data_y = event.xdata, event.ydata
        self.set_last_data_xy(data_x, data_y)

        self.make_ui_callback_viewer(self, 'motion', button,
                                     data_x, data_y)

    def key_down_event(self, event):
        key = self._get_key(event)
        modifiers = self._get_modifiers(event)
        # NOTE: event.xdata, event.ydata seem to be None
        # last_data_x, last_data_y = self.get_last_data_xy()

        self.make_ui_callback_viewer(self, 'key-press', key)

    def key_up_event(self, event):
        key = self._get_key(event)
        modifiers = self._get_modifiers(event)
        # NOTE: event.xdata, event.ydata seem to be None
        # last_data_x, last_data_y = self.get_last_data_xy()

        self.make_ui_callback_viewer(self, 'key-release', key)

    def resize_event(self, event):
        wd, ht = event.width, event.height
        self.reschedule_resize(wd, ht)

    def enter_notify_event(self, event):
        if self.t_['plot_enter_focus']:
            self.take_focus()
        self.make_callback('enter')

    def leave_notify_event(self, event):
        self.make_callback('leave')

    def focus_event(self, event, tf):
        self.make_callback('focus', tf)

    def take_focus(self):
        w = self.get_widget()
        if hasattr(w, 'setFocus'):
            # NOTE: this is a Qt call, not cross-backend
            # TODO: see if matplotlib has a backend independent way
            # to do this
            w.setFocus()
        elif hasattr(w, 'grab_focus'):
            # NOTE: this is a Gtk3 call, not cross-backend
            w.grab_focus()



#
# main.py -- reference viewer for the Ginga toolkit.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
"""This module handles the main reference viewer."""

# stdlib imports
import glob
import sys
import os
import logging
import logging.handlers
import threading

# 3rd party
import yaml

# Local application imports
from ginga.misc.Bunch import Bunch
from ginga.misc import Task, ModuleManager, Settings, log
import ginga.version as version
import ginga.toolkit as ginga_toolkit
from ginga.util import paths, rgb_cms, compat, loader, grc

# Catch warnings
logging.captureWarnings(True)

__all__ = ['ReferenceViewer']

default_layout = ['seq', {},
                   ['vbox', dict(name='top', width=1400, height=700),  # noqa
                    dict(row=['ws', dict(name='menubar', wstype='stack',
                                         group=99)],
                         stretch=0),
                    dict(row=['hpanel', dict(name='hpnl'),
                     ['ws', dict(name='left', wstype='tabs',  # noqa
                                 width=300, height=-1, group=2),
                      # (tabname, layout), ...
                      [("Info", ['vpanel', {},
                                 ['ws', dict(name='uleft', wstype='stack',
                                             height=250, group=3)],
                                 ['ws', dict(name='lleft', wstype='tabs',
                                             height=330, group=3)],
                                 ]
                        )]],
                     ['vbox', dict(name='main', width=600),
                      dict(row=['ws', dict(name='channels', wstype='tabs',
                                           group=1, use_toolbar=True,
                                           default=True)],
                           stretch=1),
                      dict(row=['ws', dict(name='cbar', wstype='stack',
                                           group=99)], stretch=0),
                      dict(row=['ws', dict(name='readout', wstype='stack',
                                           group=99)], stretch=0),
                      dict(row=['ws', dict(name='operations', wstype='stack',
                                           group=99)], stretch=0),
                      ],
                     ['ws', dict(name='right', wstype='tabs',
                                 width=400, height=-1, group=2),
                      # (tabname, layout), ...
                      [("Dialogs", ['ws', dict(name='dialogs', wstype='tabs',
                                               group=2)
                                    ]
                        )]
                      ],
                     ], stretch=1),  # noqa
                    dict(row=['ws', dict(name='toolbar', wstype='stack',
                                         height=40, group=2)],
                         stretch=0),
                    dict(row=['hbox', dict(name='status')], stretch=0),
                    ]]

plugins = [
    # hidden plugins, started at program initialization
    Bunch(module='Menubar', klass='GingaMenubar', workspace='menubar',
          start=True, hidden=True, category='System', menu="Menubar [G]",
          ptype='global', enabled=True),
    Bunch(module='Operations', workspace='operations', start=True,
          hidden=True, category='System', menu="Operations [G]",
          ptype='global', enabled=True),
    Bunch(module='Toolbar', klass='Toolbar', workspace='toolbar',
          start=True, hidden=True, category='System', ptype='global',
          enabled=True),
    Bunch(module='Toolbar', klass='Toolbar_Ginga_Image',
          hidden=True, category='System', ptype='local',
          enabled=True, exclusive=False),
    Bunch(module='Toolbar', klass='Toolbar_Ginga_Plot',
          hidden=True, category='System', ptype='local',
          enabled=True, exclusive=False),
    Bunch(module='Toolbar', klass='Toolbar_Ginga_Table',
          hidden=True, category='System', ptype='local',
          enabled=True, exclusive=False),
    Bunch(module='Pan', workspace='uleft', start=True,
          hidden=True, category='System', menu="Pan [G]", ptype='global',
          enabled=True),
    Bunch(module='Info', tab='Synopsis', workspace='lleft', start=True,
          hidden=True, category='System', menu="Info [G]", ptype='global',
          enabled=True),
    Bunch(module='Info', klass='Info_Ginga_Image',
          hidden=True, category='System', ptype='local',
          enabled=True, exclusive=False),
    Bunch(module='Info', klass='Info_Ginga_Plot',
          hidden=True, category='System', ptype='local',
          enabled=True, exclusive=False),
    Bunch(module='Info', klass='Info_Ginga_Table',
          hidden=True, category='System', ptype='local',
          enabled=True, exclusive=False),
    Bunch(module='Thumbs', tab='Thumbs', workspace='right', start=True,
          hidden=True, category='System', menu="Thumbs [G]", ptype='global',
          enabled=True),
    Bunch(module='Contents', tab='Contents', workspace='right', start=True,
          hidden=True, category='System', menu="Contents [G]", ptype='global',
          enabled=True),
    Bunch(module='Colorbar', workspace='cbar', start=True,
          hidden=True, category='System', menu="Colorbar [G]", ptype='global',
          enabled=True),
    Bunch(module='Cursor', workspace='readout', start=True,
          hidden=True, category='System', menu="Cursor [G]", ptype='global',
          enabled=True),
    Bunch(module='Errors', tab='Errors', workspace='right', start=True,
          hidden=True, category='System', menu="Errors [G]", ptype='global',
          enabled=True),
    Bunch(module='Downloads', tab='Downloads', workspace='right', start=False,
          menu="Downloads [G]", category='Utils', ptype='global', enabled=True),

    # optional, user-started plugins
    Bunch(module='Blink', tab='Blink Channels', workspace='right', start=False,
          name='Blink[channels]', menu="Blink Channels [G]",
          category='Analysis', ptype='global', enabled=True),
    Bunch(module='Blink', workspace='dialogs', menu='Blink Images',
          name='Blink[images]', category='Analysis', ptype='local',
          enabled=True),
    Bunch(module='Crosshair', workspace='left', category='Analysis',
          ptype='local', enabled=True),
    Bunch(module='Cuts', workspace='dialogs', category='Analysis',
          ptype='local', singleton=False, enabled=True),
    Bunch(module='LineProfile', workspace='dialogs',
          category='Analysis.Datacube', ptype='local', enabled=True),
    Bunch(module='Histogram', workspace='dialogs', category='Analysis',
          ptype='local', singleton=False, enabled=True),
    Bunch(module='Overlays', workspace='dialogs', category='Analysis',
          ptype='local', enabled=True),
    Bunch(module='Pick', workspace='dialogs', category='Analysis',
          ptype='local', singleton=False, enabled=True),
    Bunch(module='PixTable', workspace='dialogs', category='Analysis',
          ptype='local', enabled=True),
    Bunch(module='TVMark', workspace='dialogs', category='Analysis',
          ptype='local', enabled=True),
    Bunch(module='TVMask', workspace='dialogs', category='Analysis',
          ptype='local', enabled=True),
    Bunch(module='WCSMatch', tab='WCSMatch', workspace='right', start=False,
          menu="WCS Match [G]", category='Analysis', ptype='global',
          enabled=True),
    Bunch(module='Command', tab='Command', workspace='lleft', start=False,
          menu="Command Line [G]", category='Debug', ptype='global',
          enabled=True),
    Bunch(module='Log', tab='Log', workspace='right', start=False,
          menu="Logger Info [G]", category='Debug', ptype='global',
          enabled=True),
    Bunch(module='MultiDim', workspace='lleft', category='Navigation',
          ptype='local', exclusive=False, enabled=True),
    Bunch(module='RC', tab='RC', workspace='right', start=False,
          menu="Remote Control [G]", category='Remote', ptype='global',
          enabled=True),
    Bunch(module='SAMP', tab='SAMP', workspace='right', start=False,
          menu="SAMP Client [G]", category='Remote', ptype='global',
          enabled=False),
    Bunch(module='Compose', workspace='dialogs', category='RGB', ptype='local',
          enabled=False),
    Bunch(module='ScreenShot', workspace='dialogs', category='RGB',
          ptype='local', enabled=True),
    Bunch(module='ColorMapPicker', tab='ColorMapPicker',
          name="ColorMapPicker[G]", menu="Set Color Map [G]",
          workspace='right', start=False,
          category='RGB', ptype='global', enabled=True),
    Bunch(module='ColorMapPicker',
          menu="Set Color Map", workspace='dialogs', category='RGB',
          ptype='local', exclusive=False, enabled=True),
    Bunch(module='PlotTable', workspace='dialogs', category='Table',
          ptype='local', singleton=False, exclusive=False, enabled=True),
    Bunch(module='Catalogs', workspace='dialogs', category='Utils',
          ptype='local', enabled=True),
    Bunch(module='Drawing', workspace='dialogs', category='Utils',
          ptype='local', singleton=False, enabled=True),
    Bunch(module='AutoLoad', workspace='dialogs', category='Utils',
          ptype='local', enabled=False),
    Bunch(module='Pipeline', workspace='dialogs', category='Photos',
          ptype='local', enabled=False),
    Bunch(module='SlideShow', workspace='dialogs', category='Photos',
          ptype='local', enabled=False),
    Bunch(module='FBrowser', workspace='dialogs', category='Utils',
          ptype='local', exclusive=False, singleton=False, enabled=True),
    Bunch(module='ChangeHistory', tab='History', workspace='right',
          menu="History [G]", start=False, category='Utils', ptype='global',
          enabled=True),
    Bunch(module='Mosaic', workspace='dialogs', category='Utils', ptype='local',
          enabled=True),
    Bunch(module='Collage', workspace='dialogs', category='Utils', ptype='local',
          enabled=True),
    Bunch(module='FBrowser', tab='Open File', workspace='right',
          name="FBrowser[G]", menu="Open File [G]", start=False,
          category='Utils', ptype='global', singleton=False, enabled=True),
    Bunch(module='Preferences', workspace='dialogs', category='Utils',
          ptype='local', enabled=True),
    Bunch(module='Ruler', workspace='dialogs', category='Utils', ptype='local',
          singleton=False, enabled=True),
    # TODO: Add SaveImage to File menu.
    Bunch(module='SaveImage', tab='SaveImage', workspace='right',
          menu="Save File [G]", start=False, category='Utils', ptype='global',
          enabled=True),
    Bunch(module='WCSAxes', workspace='dialogs', category='Utils',
          ptype='local', enabled=True),
    Bunch(module='Header', tab='Header', workspace='left', start=False,
          menu="Header [G]", hidden=False, category='Utils', ptype='global',
          enabled=True),
    Bunch(module='Zoom', tab='Zoom', workspace='left', start=False,
          menu="Zoom [G]", category='Utils', ptype='global', enabled=True),
    Bunch(module='LoaderConfig', tab='Loaders', workspace='channels',
          start=False, menu="LoaderConfig [G]", category='Debug',
          ptype='global', enabled=True),
    Bunch(module='PluginConfig', tab='Plugins', workspace='channels',
          start=False, menu="PluginConfig [G]", category='Debug',
          ptype='global', enabled=True),
]


def get_plugin_spec(module=None):
    return [spec for spec in plugins if spec['module'] == module]


class ReferenceViewer(object):
    """
    This class exists solely to be able to customize the reference
    viewer startup.
    """
    def __init__(self, layout=default_layout, plugins=plugins, appname='ginga',
                 basedir=None, channels=None):
        self.appname = appname
        self.basedir = basedir
        self.layout = layout
        if channels is None:
            channels = ['Image']
        self.channels = channels
        self.default_plugins = plugins
        self.plugins = []
        self.plugin_dct = dict()

    def add_plugin_spec(self, spec):
        self.plugins.append(spec)
        plugin_name = self.get_plugin_name(spec)
        self.plugin_dct[plugin_name] = spec

    def clear_default_plugins(self):
        self.plugins = []
        self.plugin_dct = dict()

    def get_plugin_name(self, spec):
        if 'name' in spec:
            return spec['name']
        module = spec['module']
        if '.' in module:
            module = module.split('.')[-1]
        klass = spec.get('klass', None)
        name = module if klass is None else klass
        return name

    def add_default_plugins(self, except_global=[], except_local=[]):
        """
        Add the ginga-distributed default set of plugins to the
        reference viewer.
        """
        # add default plugins
        for spec in self.default_plugins:
            ptype = spec.get('ptype', 'local')
            if ptype == 'global' and spec.module not in except_global:
                self.add_plugin_spec(spec)

            if ptype == 'local' and spec.module not in except_local:
                self.add_plugin_spec(spec)

    def add_separately_distributed_plugins(self):
        groups = ['ginga.rv.plugins', 'ginga_plugins',
                  '{}_plugins'.format(self.appname)]
        available_methods = []

        for group in groups:
            discovered_plugins = compat.ep_get(group)
            for entry_point in discovered_plugins:
                try:
                    method = entry_point.load()
                    available_methods.append(method)

                except Exception as e:
                    print("Error trying to load entry point %s: %s" % (
                        str(entry_point), str(e)))

        for method in available_methods:
            try:
                spec = method()
                self.add_plugin_spec(spec)

            except Exception as e:
                print("Error trying to instantiate external plugin using %s: %s" % (
                    str(method), str(e)))

    def add_default_options(self, argprs):
        """
        Adds the default reference viewer startup options to an
        ArgumentParser instance `argprs`.
        """
        if hasattr(argprs, 'add_option'):
            # older OptParse
            add_argument = argprs.add_option
        else:
            # newer ArgParse
            add_argument = argprs.add_argument

        add_argument("--basedir", dest="basedir", metavar="NAME",
                     help="Specify Ginga configuration area")
        add_argument("--bufsize", dest="bufsize", metavar="NUM",
                     type=int, default=10,
                     help="Buffer length to NUM")
        add_argument('-c', "--channels", dest="channels",
                     help="Specify list of channels to create")
        add_argument("--disable-plugins", dest="disable_plugins",
                     metavar="NAMES",
                     help="Specify plugins that should be disabled")
        add_argument("--display", dest="display", metavar="HOST:N",
                     help="Use X display on HOST:N")
        add_argument("--fitspkg", dest="fitspkg", metavar="NAME",
                     default=None,
                     help="Prefer FITS I/O module NAME")
        add_argument("-g", "--geometry", dest="geometry",
                     default=None, metavar="GEOM",
                     help="X geometry for initial size and placement")
        add_argument("--minthreads", dest="minthreads", type=int,
                     default=None, metavar="NUM",
                     help="Start minimum of NUM threads in thread pool")
        add_argument("--modules", dest="modules", metavar="NAMES",
                     help="Specify additional modules to load")
        add_argument("--norestore", dest="norestore", default=False,
                     action="store_true",
                     help="Don't restore the GUI from a saved layout")
        add_argument("--nosplash", dest="nosplash", default=False,
                     action="store_true",
                     help="Don't display the splash screen")
        add_argument("--numthreads", dest="numthreads", type=int,
                     default=None, metavar="NUM",
                     help="Maximum NUM threads in thread pool")
        add_argument("--opengl", dest="opengl", default=False,
                     action="store_true",
                     help="Use OpenGL acceleration")
        add_argument("--plugins", dest="plugins", metavar="NAMES",
                     help="Specify additional plugins to load")
        add_argument("--rcport", dest="rc_port", type=int,
                     default=None, metavar="NUM",
                     help="Use PORT for Ginga Remote Control plugin")
        add_argument("--sep", dest="separate_channels", default=False,
                     action="store_true",
                     help="Load files in separate channels")
        add_argument("--suppress-fits-warnings",
                     dest="suppress_fits_warnings", default=False,
                     action="store_true",
                     help="Suppress FITS verify warnings")
        add_argument("-t", "--toolkit", dest="toolkit", metavar="NAME",
                     default=None,
                     help="Prefer GUI toolkit (gtk|qt)")
        add_argument("--wcspkg", dest="wcspkg", metavar="NAME",
                     default=None,
                     help="Prefer WCS module NAME")
        log.addlogopts(argprs)

    def main(self, options, args):
        """
        Main routine for running the reference viewer.

        `options` is a ArgumentParser object that has been populated with
        values from parsing the command line.  It should at least include
        the options from add_default_options()

        `args` is a list of arguments to the viewer after parsing out
        options.  It should contain a list of files or URLs to load.
        """
        logname = self.appname.lower().replace(' ', '_')

        # Create a logger
        logger = log.get_logger(name=logname, options=options)

        if options.basedir is not None:
            # command line option overrules
            self.basedir = os.path.expanduser(options.basedir)
        if self.basedir is not None:
            # custom basedir
            paths.set_home(self.basedir)
        else:
            # stock ginga basedir
            self.basedir = paths.ginga_home

        # Get settings (preferences)
        if not os.path.exists(self.basedir):
            try:
                os.mkdir(self.basedir)
            except OSError as e:
                logger.warning(
                    "Couldn't create %s settings area (%s): %s" % (
                        self.appname, self.basedir, str(e)))
                logger.warning("Preferences will not be able to be saved")

        if options.rc_port is not None:
            # user specified a custom Remote Control port
            grc.default_rc_port = options.rc_port

        # Set up preferences
        prefs = Settings.Preferences(basefolder=self.basedir, logger=logger)
        settings = prefs.create_category('general')
        settings.set_defaults(appname=self.appname,
                              title=self.appname.capitalize(),
                              useMatplotlibColormaps=False,
                              widgetSet='choose',
                              WCSpkg='choose', FITSpkg='choose',
                              suppress_fits_warnings=False,
                              recursion_limit=2000,
                              min_threads=2,
                              num_threads=max(os.cpu_count(), 10),
                              threadpool_analyze_interval_sec=None,
                              pluginmgr_allow_nonsingletons=True,
                              icc_working_profile=None,
                              font_scaling_factor=None,
                              save_layout=True,
                              use_opengl=False,
                              layout_file='layout.json',
                              plugin_file='plugins.yml',
                              channel_prefix="Image")
        settings.load(onError='silent')

        # default of 1000 is a little too small
        sys.setrecursionlimit(settings.get('recursion_limit'))

        # So we can find our plugins
        sys.path.insert(0, self.basedir)
        package_home = os.path.split(sys.modules['ginga.version'].__file__)[0]
        child_dir = os.path.join(package_home, 'rv', 'plugins')
        sys.path.insert(0, child_dir)
        plugin_dir = os.path.join(self.basedir, 'plugins')
        sys.path.insert(0, plugin_dir)

        # Create the dynamic module manager
        mm = ModuleManager.ModuleManager(logger)
        sys.meta_path.append(mm)

        # what is the dynamic config file to load
        app_config = "{}_config".format(self.appname)
        gc = os.path.join(self.basedir, f"{app_config}.py")
        have_app_config = os.path.exists(gc)

        # User configuration, earliest possible intervention
        if have_app_config:
            try:
                app_config = mm.load_module(app_config)

                if hasattr(app_config, 'init_config'):
                    app_config.init_config(self)

            except Exception as e:
                logger.error("Error processing %s config file: %s" % (
                    self.appname, str(e)), exc_info=True)

        # Choose a toolkit
        if options.toolkit:
            toolkit = options.toolkit
        else:
            toolkit = settings.get('widgetSet', 'choose')

        if toolkit == 'choose':
            try:
                ginga_toolkit.choose()
            except ImportError as e:
                print("UI toolkit choose error: %s" % str(e))
                sys.exit(1)
        else:
            ginga_toolkit.use(toolkit)

        tkname = ginga_toolkit.get_family()
        logger.info("Chosen toolkit (%s) family is '%s'" % (
            ginga_toolkit.toolkit, tkname))

        # these imports have to be here, otherwise they force the choice
        # of toolkit too early
        from ginga.rv.Control import GingaShell, GuiLogHandler

        if settings.get('useMatplotlibColormaps', False):
            # Add matplotlib color maps if matplotlib is installed
            try:
                from ginga import cmap
                cmap.add_matplotlib_cmaps(fail_on_import_error=False)
            except Exception as e:
                logger.warning(
                    "failed to load matplotlib colormaps: %s" % (str(e)))

        # Set a working RGB ICC profile if user has one
        working_profile = settings.get('icc_working_profile', None)
        rgb_cms.working_profile = working_profile

        # User wants to customize the WCS package?
        if options.wcspkg:
            wcspkg = options.wcspkg
        else:
            wcspkg = settings.get('WCSpkg', 'choose')

        try:
            from ginga.util import wcsmod
            if wcspkg != 'choose':
                assert wcsmod.use(wcspkg) is True
        except Exception as e:
            logger.warning(
                "failed to set WCS package preference '{}': {}".format(wcspkg, e))

        # User wants to customize the FITS package?
        if options.fitspkg:
            fitspkg = options.fitspkg
        else:
            fitspkg = settings.get('FITSpkg', 'choose')

        if options.suppress_fits_warnings:
            supp_warn = options.suppress_fits_warnings
        else:
            supp_warn = settings.get('suppress_fits_warnings', False)
        if supp_warn:
            import warnings
            from astropy.io import fits
            warnings.simplefilter('ignore', fits.verify.VerifyWarning)

        try:
            from ginga.util.io import io_fits
            if fitspkg != 'choose':
                assert io_fits.use(fitspkg) is True

        except Exception as e:
            logger.warning(
                "failed to set FITS package preference '{}': {}".format(fitspkg, e))

        ev_quit = threading.Event()
        # Create and start thread pool
        num_threads = settings.get('num_threads', max(os.cpu_count(), 10))
        if options.numthreads is not None:
            num_threads = options.numthreads
        min_threads = settings.get('min_threads', 2)
        if options.minthreads is not None:
            min_threads = options.minthreads
        analyze_interval = settings.get('threadpool_analyze_interval_sec', None)
        thread_pool = Task.ThreadPool(numthreads=num_threads, logger=logger,
                                      minthreads=min_threads, ev_quit=ev_quit,
                                      analyze_interval=analyze_interval)
        thread_pool.startall()

        # Create the Ginga main object
        ginga_shell = GingaShell(logger, thread_pool, mm, prefs,
                                 ev_quit=ev_quit)

        if options.opengl:
            settings.set(use_opengl=True)

        layout_file = os.path.join(self.basedir, settings.get('layout_file',
                                                              'layout.json'))
        ginga_shell.set_layout(self.layout, layout_file=layout_file,
                               save_layout=settings.get('save_layout', True))

        # User configuration (custom star catalogs, etc.)
        if have_app_config:
            try:
                if hasattr(app_config, 'pre_gui_config'):
                    app_config.pre_gui_config(ginga_shell)
            except Exception as e:
                logger.error("Error importing %s config file: %s" % (
                    self.appname, str(e)), exc_info=True)

        # Build desired layout
        ginga_shell.build_toplevel(ignore_saved_layout=options.norestore)

        # Did user specify a particular geometry?
        if options.geometry:
            ginga_shell.set_geometry(options.geometry)

        # make the list of disabled plugins
        if options.disable_plugins is not None:
            disabled_plugins = options.disable_plugins.lower().split(',')
        else:
            disabled_plugins = settings.get('disable_plugins', [])
            if not isinstance(disabled_plugins, list):
                disabled_plugins = disabled_plugins.lower().split(',')
        disabled_plugins = set(disabled_plugins)

        # Add GUI log handler (for "Log" global plugin)
        guiHdlr = GuiLogHandler(ginga_shell)
        guiHdlr.setLevel(options.loglevel)
        fmt = logging.Formatter(log.LOG_FORMAT)
        guiHdlr.setFormatter(fmt)
        logger.addHandler(guiHdlr)

        # Set loader priorities, if user has saved any
        # (see LoaderConfig plugin)
        path = os.path.join(self.basedir, 'loaders.yml')
        if os.path.exists(path):
            try:
                with open(path, 'r') as in_f:
                    loader_dct = yaml.safe_load(in_f.read())

                # set saved priorities for openers
                for mimetype, m_dct in loader_dct.items():
                    for name, l_dct in m_dct.items():
                        opener = loader.get_opener(name)
                        loader.add_opener(opener, [mimetype],
                                          priority=l_dct['priority'],
                                          note=opener.__doc__)

            except Exception as e:
                logger.error(f"failed to process loader file '{path}': {e}",
                             exc_info=True)

        # Does user have a saved plugin setup?  If so, check which
        # plugins should be disabled, or have a customized category or
        # workspace
        plugin_file = settings.get('plugin_file', None)
        if plugin_file is not None:
            plugin_file = os.path.join(self.basedir, plugin_file)
            if os.path.exists(plugin_file):
                logger.info("Reading plugin file '%s'..." % (plugin_file))
                try:
                    with open(plugin_file, 'r') as in_f:
                        buf = in_f.read()
                        _plugins = yaml.safe_load(buf)

                    for dct in _plugins:
                        plugin_name = self.get_plugin_name(dct)
                        if plugin_name in self.plugin_dct:
                            # we know about this plugin, override listed
                            # attributes
                            spec = self.plugin_dct[plugin_name]
                            for name in ['enabled', 'category', 'hidden',
                                         'workspace', 'tab', 'menu', 'start',
                                         'pfx', 'optray', 'singleton', 'limit',
                                         'exclusive']:
                                if name in dct:
                                    spec[name] = dct[name]
                        else:
                            # unknown plugin
                            spec = Bunch(dct)
                            self.add_plugin_spec(spec)

                except Exception as e:
                    logger.error(f"Error reading plugin file: {e}",
                                 exc_info=True)

        # Load any custom global plugins named on command line or in
        # general.cfg
        if options.modules is not None:
            global_plugins = options.modules.split(',')
        else:
            global_plugins = settings.get('global_plugins', [])
            if not isinstance(global_plugins, list):
                global_plugins = global_plugins.split(',')

        for long_plugin_name in global_plugins:
            if '.' in long_plugin_name:
                tmpstr = long_plugin_name.split('.')
                plugin_name = tmpstr[-1]
                pfx = '.'.join(tmpstr[:-1])
            else:
                plugin_name = long_plugin_name
                pfx = None
            if plugin_name in self.plugin_dct:
                spec = self.plugin_dct[plugin_name]
                spec.enabled = True
                spec.start = True
            else:
                menu_name = f"{plugin_name} [G]"
                spec = Bunch(name=plugin_name, module=plugin_name,
                             ptype='global', tab=plugin_name,
                             menu=menu_name, category="Custom",
                             enabled=True, start=True,
                             workspace='right', pfx=pfx)
                self.add_plugin_spec(spec)

        # Load any custom local plugins named on command line or in
        # general.cfg
        if options.plugins is not None:
            local_plugins = options.plugins.split(',')
        else:
            local_plugins = settings.get('local_plugins', [])
            if not isinstance(local_plugins, list):
                local_plugins = local_plugins.split(',')

        for long_plugin_name in local_plugins:
            if '.' in long_plugin_name:
                tmpstr = long_plugin_name.split('.')
                plugin_name = tmpstr[-1]
                pfx = '.'.join(tmpstr[:-1])
            else:
                plugin_name = long_plugin_name
                pfx = None
            if plugin_name in self.plugin_dct:
                spec = self.plugin_dct[plugin_name]
                spec.enabled = True
            else:
                spec = Bunch(module=plugin_name, workspace='dialogs',
                             ptype='local', category="Custom",
                             hidden=False, pfx=pfx, enabled=True)
                self.add_plugin_spec(spec)

        # Mark disabled plugins (command-line has precedence)
        for spec in self.plugins:
            if spec.module.lower() in disabled_plugins:
                spec['enabled'] = False

        # submit plugin specs to shell
        ginga_shell.set_plugins(self.plugins)

        # start any plugins that have start=True
        ginga_shell.boot_plugins()
        ginga_shell.update_pending()

        # TEMP?
        tab_names = [name.lower()
                     for name in ginga_shell.ds.get_tabnames(group=None)]
        if 'info' in tab_names:
            ginga_shell.ds.raise_tab('Info')
        if 'synopsis' in tab_names:
            ginga_shell.ds.raise_tab('Synopsis')
        if 'thumbs' in tab_names:
            ginga_shell.ds.raise_tab('Thumbs')

        # Add custom channels
        if options.channels is not None:
            channels = options.channels.split(',')
        else:
            channels = settings.get('channels', self.channels)
            if not isinstance(channels, list):
                channels = channels.split(',')

        if len(channels) > 0:
            # populate the initial channel lineup
            for item in channels:
                if isinstance(item, str):
                    chname, wsname = item, None
                else:
                    chname, wsname = item
                ginga_shell.add_channel(chname, workspace=wsname)

            ginga_shell.change_channel(chname)

        # User configuration (custom star catalogs, etc.)
        if have_app_config:
            try:
                if hasattr(app_config, 'post_gui_config'):
                    app_config.post_gui_config(ginga_shell)

            except Exception as e:
                logger.error("Error processing %s config file: %s" % (
                    self.appname, str(e)), exc_info=True)

        # Redirect warnings to logger
        for hdlr in logger.handlers:
            logging.getLogger('py.warnings').addHandler(hdlr)

        # Display banner the first time run, unless suppressed
        show_banner = (self.appname == 'ginga')
        try:
            show_banner = settings.get('showBanner')

        except KeyError:
            # disable for subsequent runs
            settings.set(showBanner=False)
            if not os.path.exists(settings.preffile):
                settings.save()

        if (not options.nosplash) and (len(args) == 0) and show_banner:
            ginga_shell.banner()

        # Handle inputs like "*.fits[ext]" that sys cmd cannot auto expand.
        expanded_args = []
        for imgfile in args:
            if '*' in imgfile:
                if '[' in imgfile and imgfile.endswith(']'):
                    s = imgfile.split('[')
                    ext = '[' + s[1]
                    imgfile = s[0]
                else:
                    ext = ''
                for fname in glob.iglob(imgfile):
                    expanded_args.append(fname + ext)
            else:
                expanded_args.append(imgfile)

        if len(channels) > 0:
            # Assume remaining arguments are fits files and load them.
            if not options.separate_channels:
                chname = channels[0]
                ginga_shell.gui_do(ginga_shell.open_uris, expanded_args,
                                   chname=chname)
            else:
                i = 0
                num_channels = len(channels)
                for imgfile in expanded_args:
                    if i < num_channels:
                        chname = channels[i]
                        i = i + 1
                    else:
                        channel = ginga_shell.add_channel_auto()
                        chname = channel.name
                    ginga_shell.gui_do(ginga_shell.open_uris, [imgfile],
                                       chname=chname)

        try:
            try:
                # if there is a network component, start it
                if hasattr(ginga_shell, 'start'):
                    logger.info("starting network interface...")
                    ginga_shell.start()

                # Main loop to handle GUI events
                logger.info("entering mainloop...")
                ginga_shell.mainloop(timeout=0.001)

            except KeyboardInterrupt:
                logger.error("Received keyboard interrupt!")

        finally:
            logger.info("Shutting down...")
            ev_quit.set()

        sys.exit(0)


def reference_viewer(sys_argv):
    """Create reference viewer from command line."""
    viewer = ReferenceViewer(layout=default_layout, plugins=plugins,
                             appname='ginga', basedir=None, channels=['Image'])
    viewer.add_default_plugins()
    viewer.add_separately_distributed_plugins()

    # Parse command line options with argparse module
    from argparse import ArgumentParser

    argprs = ArgumentParser(description="Run the Ginga reference viewer.")
    viewer.add_default_options(argprs)
    argprs.add_argument('-V', '--version', action='version',
                        version='%(prog)s {}'.format(version.version))
    (options, args) = argprs.parse_known_args(sys_argv[1:])

    if options.display:
        os.environ['DISPLAY'] = options.display

    viewer.main(options, args)


def _main():
    """Run from command line."""
    reference_viewer(sys.argv)

# END



#
# mosaic.py -- Classes for quick and dirty mosaicing of FITS images
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import math
import time
import warnings

import numpy as np

from ginga import AstroImage, trcalc
from ginga.util import wcs, loader, dp, iqcalc
from ginga.util.io import io_fits
from ginga.misc import Callback, Settings


def get_warp_indexes(shape_in, wcs_in, wcs_out):
    """Get numpy index arrays to warp an image into another projection.

    For every pixel coordinate for numpy array shape ``shape_in``,
    convert to wcs coordinates according to ``wcs_in`` and then back
    to pixel coordinates according to ``wcs_out``.

    Parameters
    ----------
    shape_in : numpy ndarray shape
        The shape of the array to be projected

    wcs_in : subclass of `~ginga.util.wcsmod.common.BaseWCS`
        Ginga WCS wrapper that is associated with the input array

    wcs_out : subclass of `~ginga.util.wcsmod.common.BaseWCS`
        Ginga WCS wrapper that is associated with the output

    Returns
    -------
    Returns a 3-tuple (old_pts, coords, new_pts), where all three
    arrays are Nx2
    """
    ht, wd = shape_in[:2]
    yi, xi = np.mgrid[0:ht, 0:wd]
    old_pts = np.array((xi.ravel(), yi.ravel())).T

    # convert data coords of pixels as sky coords
    coords = wcs_in.datapt_to_wcspt(old_pts)

    # calc these sky points in data x, y according to the *wcs_ref*
    new_pts = wcs_out.wcspt_to_datapt(coords)

    return (old_pts, coords, new_pts)


def warp_image(data_in, wcs_in, wcs_out, fill=None, pixel_radius=1):
    """Warp image in 2D numpy array ``data`` to a new array.

    Warps ``data_in`` using ``wcs_in`` and projecting into ``wcs_out``.
    The resulting array may have empty pixels which are initially filled
    with ``fill`` and then computed from the median value of surrounding
    pixels at a radius of ``pixel_radius``.

    Parameters
    ----------
    data_in : numpy 2D ndarray
        The array to be warped (projected)

    wcs_in : subclass of `~ginga.util.wcsmod.common.BaseWCS`
        Ginga WCS wrapper that is associated with the input array

    wcs_out : subclass of `~ginga.util.wcsmod.common.BaseWCS`
        Ginga WCS wrapper that is associated with the output

    fill : scalar value or `None` (optional, defaults to `None`)
        The value to initially fill the new output image array

    pixel_radius : `int` (optional, defaults to 1)
        The pixel radius to use for collecting values to fill empty pixels

    Returns
    -------
    Returns a 5-tuple (data_out, old_pts, coords, new_pts, out_pts).
    ``data_out`` is the warped image, with an alpha mask layer attached.
    ``out_pts`` is a Nx2 array describing the relocated pixels in data_out.

    See ``get_warp_indexes`` for discussion of the ``old_pts``, ``coords``
    and ``new_pts`` return values.
    """
    old_pts, coords, new_pts = get_warp_indexes(data_in.shape, wcs_in, wcs_out)

    # round to nearest int
    new_pts = np.rint(new_pts).astype(int)

    # get bounds of array values
    mn, mx = trcalc.get_bounds(new_pts)

    # subtract minimums to turn pixel coordinates into offsets
    out_pts = new_pts - mn

    # np fancy indexing to warp array as necessary
    x, y = old_pts.T
    nx, ny = out_pts.T

    if fill is None:
        # select a suitable fill value if one is not provided
        if issubclass(data_in.dtype.type, np.floating):
            fill = np.nan
        else:
            fill = 0

    # allocate new array
    new_wd, new_ht = mx - mn + np.array((1, 1))
    data_out = np.full((new_ht, new_wd, 2), fill, dtype=data_in.dtype)

    # prepare mask, will be True where there are empty pixels in the image
    mask = np.full(data_out.shape[:2], True, dtype=bool)
    mask[ny, nx] = False

    # fill alpha layer with zeros where we have empty pixels and ones
    # otherwise
    data_out[..., 1][mask] = 0
    data_out[..., 1][~mask] = 1

    # warp the data into the destination image
    data_out[ny, nx, 0] = data_in[y, x]

    if pixel_radius > 0:
        # fill in holes in output image with median values of surrounding
        # pixels  NOTE: this also fills in alpha layer values
        y, x = np.where(mask)
        pr = pixel_radius
        offsets = [(x, y)
                   for x in range(-pr, pr + 1) for y in range(-pr, pr + 1)]
        offsets.remove((0, 0))
        _arrs = [data_out[_y, _x]
                 for _x, _y in [((x + offsets[i][0]).clip(0, new_wd - 1),
                                 (y + offsets[i][1]).clip(0, new_ht - 1))
                                for i in range(len(offsets))]]

        with warnings.catch_warnings():
            # we can get a "RuntimeWarning: All-NaN slice encountered"
            # which is ok as we simply let this resolve to NaN
            warnings.simplefilter("ignore")
            vals = np.nanmedian(np.dstack(_arrs), axis=2)
        _arrs = None   # deref arrays
        data_out[y, x] = vals

    # finally multipy alpha layer by mx_v to achieve full opacity where it
    # is wanted
    mn_v, mx_v = trcalc.get_minmax_dtype(data_out.dtype)
    data_out[..., 1] *= mx_v

    return (data_out, old_pts, coords, new_pts, out_pts)


def mosaic_inline(baseimage, imagelist, bg_ref=None, trim_px=None,
                  merge=False, allow_expand=True, expand_pad_deg=0.01,
                  max_expand_pct=None,
                  update_minmax=True, suppress_callback=False):
    """Drops new images into the image `baseimage` (if there is room),
    relocating them according the WCS between the two images.
    """
    # Get our own (mosaic) rotation and scale
    header = baseimage.get_header()
    ((xrot_ref, yrot_ref),
     (cdelt1_ref, cdelt2_ref)) = wcs.get_xy_rotation_and_scale(header)

    scale_x, scale_y = math.fabs(cdelt1_ref), math.fabs(cdelt2_ref)

    # drop each image in the right place in the new data array
    mydata = baseimage._get_data()

    count = 1
    res = []
    for image in imagelist:
        name = image.get('name', 'image%d' % (count))
        count += 1

        data_np = image._get_data()
        if 0 in data_np.shape:
            baseimage.logger.info("Skipping image with zero length axis")
            continue

        # Calculate sky position at the center of the piece
        ctr_x, ctr_y = trcalc.get_center(data_np)
        ra, dec = image.pixtoradec(ctr_x, ctr_y, coords='data')

        # User specified a trim?  If so, trim edge pixels from each
        # side of the array
        ht, wd = data_np.shape[:2]
        if trim_px:
            xlo, xhi = trim_px, wd - trim_px
            ylo, yhi = trim_px, ht - trim_px
            data_np = data_np[ylo:yhi, xlo:xhi, ...]
            ht, wd = data_np.shape[:2]

        # If caller asked us to match background of pieces then
        # get the median of this piece
        if bg_ref is not None:
            bg = iqcalc.get_median(data_np)
            bg_inc = bg_ref - bg
            data_np = data_np + bg_inc

        # Determine max/min to update our values
        if update_minmax:
            maxval = np.nanmax(data_np)
            minval = np.nanmin(data_np)
            baseimage.maxval = max(baseimage.maxval, maxval)
            baseimage.minval = min(baseimage.minval, minval)

        # Get rotation and scale of piece
        header = image.get_header()
        ((xrot, yrot),
         (cdelt1, cdelt2)) = wcs.get_xy_rotation_and_scale(header)
        baseimage.logger.debug("image(%s) xrot=%f yrot=%f cdelt1=%f "
                               "cdelt2=%f" % (name, xrot, yrot, cdelt1, cdelt2))

        # scale if necessary
        # TODO: combine with rotation?
        if (not np.isclose(math.fabs(cdelt1), scale_x) or
            not np.isclose(math.fabs(cdelt2), scale_y)):
            nscale_x = math.fabs(cdelt1) / scale_x
            nscale_y = math.fabs(cdelt2) / scale_y
            baseimage.logger.debug("scaling piece by x(%f), y(%f)" % (
                nscale_x, nscale_y))
            data_np, (ascale_x, ascale_y) = trcalc.get_scaled_cutout_basic(
                data_np, 0, 0, wd - 1, ht - 1, nscale_x, nscale_y,
                logger=baseimage.logger)

        # Rotate piece into our orientation, according to wcs
        rot_dx, rot_dy = xrot - xrot_ref, yrot - yrot_ref

        flip_x = False
        flip_y = False

        # Optomization for 180 rotations
        if (np.isclose(math.fabs(rot_dx), 180.0) or
            np.isclose(math.fabs(rot_dy), 180.0)):
            rotdata = trcalc.transform(data_np,
                                       flip_x=True, flip_y=True)
            rot_dx = 0.0
            rot_dy = 0.0
        else:
            rotdata = data_np

        # Finish with any necessary rotation of piece
        if not np.isclose(rot_dy, 0.0):
            rot_deg = rot_dy
            baseimage.logger.debug("rotating %s by %f deg" % (name, rot_deg))
            rotdata = trcalc.rotate(rotdata, rot_deg,
                                    #rotctr_x=ctr_x, rotctr_y=ctr_y
                                    logger=baseimage.logger)

        # Flip X due to negative CDELT1
        if np.sign(cdelt1) != np.sign(cdelt1_ref):
            flip_x = True

        # Flip Y due to negative CDELT2
        if np.sign(cdelt2) != np.sign(cdelt2_ref):
            flip_y = True

        if flip_x or flip_y:
            rotdata = trcalc.transform(rotdata,
                                       flip_x=flip_x, flip_y=flip_y)

        # Get size and data of new image
        ht, wd = rotdata.shape[:2]
        ctr_x, ctr_y = trcalc.get_center(rotdata)

        # Find location of image piece (center) in our array
        x0, y0 = baseimage.radectopix(ra, dec, coords='data')

        # Merge piece as closely as possible into our array
        # Unfortunately we lose a little precision rounding to the
        # nearest pixel--can't be helped with this approach
        x0, y0 = int(np.rint(x0)), int(np.rint(y0))
        baseimage.logger.debug("Fitting image '%s' into mosaic at %d,%d" % (
            name, x0, y0))

        # This is for useful debugging info only
        my_ctr_x, my_ctr_y = trcalc.get_center(mydata)
        off_x, off_y = x0 - my_ctr_x, y0 - my_ctr_y
        baseimage.logger.debug("centering offsets: %d,%d" % (off_x, off_y))

        # Sanity check piece placement
        xlo, xhi = x0 - ctr_x, x0 + wd - ctr_x
        ylo, yhi = y0 - ctr_y, y0 + ht - ctr_y
        assert (xhi - xlo == wd), \
            Exception("Width differential %d != %d" % (xhi - xlo, wd))
        assert (yhi - ylo == ht), \
            Exception("Height differential %d != %d" % (yhi - ylo, ht))

        mywd, myht = baseimage.get_size()
        if xlo < 0 or xhi > mywd or ylo < 0 or yhi > myht:
            if not allow_expand:
                raise Exception("New piece doesn't fit on image and "
                                "allow_expand=False")

            # <-- Resize our data array to allow the new image

            # determine amount to pad expansion by
            expand_x = max(int(expand_pad_deg / scale_x), 0)
            expand_y = max(int(expand_pad_deg / scale_y), 0)

            nx1_off, nx2_off = 0, 0
            if xlo < 0:
                nx1_off = abs(xlo) + expand_x
            if xhi > mywd:
                nx2_off = (xhi - mywd) + expand_x
            xlo, xhi = xlo + nx1_off, xhi + nx1_off

            ny1_off, ny2_off = 0, 0
            if ylo < 0:
                ny1_off = abs(ylo) + expand_y
            if yhi > myht:
                ny2_off = (yhi - myht) + expand_y
            ylo, yhi = ylo + ny1_off, yhi + ny1_off

            new_wd = mywd + nx1_off + nx2_off
            new_ht = myht + ny1_off + ny2_off

            # sanity check on new mosaic size
            old_area = mywd * myht
            new_area = new_wd * new_ht
            expand_pct = new_area / old_area
            if ((max_expand_pct is not None) and
                    (expand_pct > max_expand_pct)):
                raise Exception("New area exceeds current one by %.2f %%;"
                                "increase max_expand_pct (%.2f) to allow" %
                                (expand_pct * 100, max_expand_pct))

            # go for it!
            new_data = np.zeros((new_ht, new_wd))
            # place current data into new data
            new_data[ny1_off:ny1_off + myht, nx1_off:nx1_off + mywd] = \
                mydata
            baseimage._data = new_data
            mydata = new_data

            if (nx1_off > 0) or (ny1_off > 0):
                # Adjust our WCS for relocation of the reference pixel
                crpix1, crpix2 = baseimage.get_keywords_list('CRPIX1', 'CRPIX2')
                kwds = dict(CRPIX1=crpix1 + nx1_off,
                            CRPIX2=crpix2 + ny1_off,
                            NAXIS1=new_wd, NAXIS2=new_ht)
                baseimage.update_keywords(kwds)

        # fit image piece into our array
        try:
            if merge:
                mydata[ylo:yhi, xlo:xhi, ...] += rotdata[0:ht, 0:wd, ...]
            else:
                idx = (mydata[ylo:yhi, xlo:xhi, ...] == 0.0)
                mydata[ylo:yhi, xlo:xhi, ...][idx] = \
                    rotdata[0:ht, 0:wd, ...][idx]

        except Exception as e:
            baseimage.logger.error("Error fitting tile: %s" % (str(e)))
            raise

        res.append((xlo, ylo, xhi, yhi))

    # TODO: recalculate min and max values
    # Can't use usual techniques because it adds too much time to the
    # mosacing
    #baseimage._set_minmax()

    # Notify watchers that our data has changed
    if not suppress_callback:
        baseimage.make_callback('modified')

    return res


def mosaic(logger, itemlist, fov_deg=None):
    """
    Parameters
    ----------
    logger : logger object
        a logger object passed to created AstroImage instances
    itemlist : sequence like
        a sequence of either filenames or AstroImage instances
    """

    if isinstance(itemlist[0], AstroImage.AstroImage):
        image0 = itemlist[0]
        name = image0.get('name', 'image0')
    else:
        # Assume it is a file and load it
        filepath = itemlist[0]
        logger.info("Reading file '%s' ..." % (filepath))
        image0 = loader.load_data(filepath, logger=logger)
        name = filepath

    ra_deg, dec_deg = image0.get_keywords_list('CRVAL1', 'CRVAL2')
    header = image0.get_header()
    (rot_deg, cdelt1, cdelt2) = wcs.get_rotation_and_scale(header)
    logger.debug("image0 rot=%f cdelt1=%f cdelt2=%f" % (rot_deg,
                                                        cdelt1, cdelt2))

    px_scale = math.fabs(cdelt1)
    expand = False
    if fov_deg is None:
        # TODO: calculate fov?
        expand = True

    cdbase = [np.sign(cdelt1), np.sign(cdelt2)]
    img_mosaic = dp.create_blank_image(ra_deg, dec_deg,
                                       fov_deg, px_scale, rot_deg,
                                       cdbase=cdbase,
                                       logger=logger)
    header = img_mosaic.get_header()
    (rot, cdelt1, cdelt2) = wcs.get_rotation_and_scale(header)
    logger.debug("mosaic rot=%f cdelt1=%f cdelt2=%f" % (rot, cdelt1, cdelt2))

    logger.debug("Processing '%s' ..." % (name))
    tup = mosaic_inline(img_mosaic, [image0], allow_expand=expand)
    logger.debug("placement %s" % (str(tup)))

    count = 1
    for item in itemlist[1:]:
        if isinstance(item, AstroImage.AstroImage):
            image = item
        else:
            # Create and load the image
            filepath = item
            logger.info("Reading file '%s' ..." % (filepath))
            image = io_fits.load_file(filepath, logger=logger)

        name = image.get('name', 'image%d' % (count))

        logger.debug("Inlining '%s' ..." % (name))
        tup = mosaic_inline(img_mosaic, [image])
        logger.debug("placement %s" % (str(tup)))
        count += 1

    logger.info("Done.")
    return img_mosaic


class ImageMosaicer(Callback.Callbacks):
    """Class for creating mosaics in a `~ginga.AstroImage.AstroImage`.

    Individual tiles are transformed and inserted into the right place
    in the image ndarray.  The array can be automatically enlarged as
    necessary to accomodate the new tiles.

    Typical usage:

    >>> mosaicer = ImageMosaicer(logger)
    >>> mosaicer.mosaic(images)

    where ``images`` is a list of `~ginga.AstroImage.AstroImage` that
    should be plotted in ``viewer``.
    """
    def __init__(self, logger, settings=None):
        super(ImageMosaicer, self).__init__()

        self.logger = logger

        self.ingest_count = 0
        # holds processed images to be inserted into mosaic image
        self.total_images = 0
        self.image_list = []

        # options
        if settings is None:
            settings = Settings.SettingGroup(name='mosaicer',
                                             logger=self.logger)
        self.t_ = settings
        self.t_.set_defaults(fov_deg=0.2,
                             match_bg=False, trim_px=0, merge=False,
                             mosaic_hdus=False, skew_limit=0.1,
                             allow_expand=True, expand_pad_deg=0.01,
                             reuse_image=False, mosaic_method='simple',
                             update_minmax=True, max_expand_pct=None,
                             annotate_images=False, annotate_color='pink',
                             annotate_fontsize=10.0, ann_fits_kwd=None,
                             ann_tag_pfx='ann_')

        # these are updated in prepare_mosaic() and represent measurements
        # on the reference image
        self.bg_ref = 0.0
        self.xrot_ref, self.yrot_ref = 0.0, 0.0
        self.cdelt1_ref, self.cdelt2_ref = 1.0, 1.0
        self.scale_x = 1.0
        self.scale_y = 1.0
        self.baseimage = None

        for name in ['progress', 'finished']:
            self.enable_callback(name)

    def get_settings(self):
        return self.t_

    def prepare_mosaic(self, ref_image):
        """Prepare a new (blank) mosaic image based on the pointing of
        the reference image ``ref_image``.
        Returns a `~ginga.AstroImage.AstroImage`

        This method is typically called internally.
        """
        # if user requesting us to match backgrounds, then calculate
        # median of root image and save it
        data_np = ref_image.get_data()
        dtype = data_np.dtype
        #dtype = None
        if issubclass(dtype.type, np.floating):
            fill = np.nan
        else:
            fill = 0
        fov_deg = self.t_['fov_deg']

        if self.t_['match_bg']:
            self.bg_ref = iqcalc.get_median(data_np)

        header = ref_image.get_header()
        ra_deg, dec_deg = header['CRVAL1'], header['CRVAL2']

        (rot_deg, cdelt1, cdelt2) = wcs.get_rotation_and_scale(header,
                                                               skew_threshold=self.t_['skew_limit'])
        self.logger.debug("ref_image rot=%f cdelt1=%f cdelt2=%f" % (rot_deg,
                                                                    cdelt1, cdelt2))

        # Prepare pixel scale for each axis
        px_scale = (math.fabs(cdelt1), math.fabs(cdelt2))
        cdbase = [np.sign(cdelt1), np.sign(cdelt2)]

        if not self.t_['reuse_image'] or self.baseimage is None:
            self.logger.debug("creating blank image to hold mosaic")
            # GC old mosaic
            self.baseimage = None

            self.baseimage = dp.create_blank_image(ra_deg, dec_deg,
                                                   fov_deg, px_scale, rot_deg,
                                                   cdbase=cdbase,
                                                   logger=self.logger,
                                                   pfx='mosaic',
                                                   dtype=dtype,
                                                   alpha=0, fill=fill)
        else:
            # <-- reuse image (faster)
            self.logger.debug("Reusing previous mosaic image")
            dp.recycle_image(self.baseimage, ra_deg, dec_deg,
                             fov_deg, px_scale, rot_deg,
                             cdbase=cdbase,
                             logger=self.logger,
                             pfx='mosaic', alpha=0, fill=fill)

        header = self.baseimage.get_header()

        # TODO: handle skew (differing rotation for each axis)?
        (rot_xy, cdelt_xy) = wcs.get_xy_rotation_and_scale(header)
        self.logger.debug("ref image rot_x=%f rot_y=%f cdelt1=%f cdelt2=%f" % (
            rot_xy[0], rot_xy[1], cdelt_xy[0], cdelt_xy[1]))

        # Store base image rotation and scale
        self.xrot_ref, self.yrot_ref = rot_xy
        self.cdelt1_ref, self.cdelt2_ref = cdelt_xy
        self.scale_x = math.fabs(cdelt_xy[0])
        self.scale_y = math.fabs(cdelt_xy[1])

        return self.baseimage

    def ingest_image(self, image):
        """Ingest an image, transform it and merge it in the right place in
        the image array.

        This method is typically called internally.
        """
        self.ingest_count += 1
        count = self.ingest_count
        tag = 'image{}'.format(count)
        name = image.get('name', tag)

        data_np = image._get_data()
        if 0 in data_np.shape:
            self.logger.info("Skipping image with zero length axis")
            return

        # Calculate sky position at the center of the piece
        ctr_x, ctr_y = trcalc.get_center(data_np)
        ra, dec = image.pixtoradec(ctr_x, ctr_y, coords='data')
        self.image_list.append((name, tag, ra, dec))

        # User specified a trim?  If so, trim edge pixels from each
        # side of the array
        ht, wd = data_np.shape[:2]
        if self.t_['trim_px'] is not None:
            xlo, xhi = self.t_['trim_px'], wd - self.t_['trim_px']
            ylo, yhi = self.t_['trim_px'], ht - self.t_['trim_px']
            data_np = data_np[ylo:yhi, xlo:xhi, ...]
            ht, wd = data_np.shape[:2]

        # If caller asked us to match background of pieces then
        # fix up this data
        if self.t_['match_bg']:
            bg = iqcalc.get_median(data_np)
            bg_inc = self.bg_ref - bg
            data_np = data_np + bg_inc

        # Determine max/min to update our values
        if self.t_['update_minmax']:
            maxval = np.nanmax(data_np)
            minval = np.nanmin(data_np)
            self.baseimage.maxval = max(self.baseimage.maxval, maxval)
            self.baseimage.minval = min(self.baseimage.minval, minval)

        # Get rotation and scale of piece
        header = image.get_header()
        ((xrot, yrot),
         (cdelt1, cdelt2)) = wcs.get_xy_rotation_and_scale(header)
        self.logger.debug("image(%s) xrot=%f yrot=%f cdelt1=%f "
                          "cdelt2=%f" % (name, xrot, yrot, cdelt1, cdelt2))

        # scale if necessary to scale of reference image
        if (not np.isclose(math.fabs(cdelt1), self.scale_x) or
            not np.isclose(math.fabs(cdelt2), self.scale_y)):
            nscale_x = math.fabs(cdelt1) / self.scale_x
            nscale_y = math.fabs(cdelt2) / self.scale_y
            self.logger.debug("scaling piece by x(%f), y(%f)" % (
                nscale_x, nscale_y))
            data_np, (ascale_x, ascale_y) = trcalc.get_scaled_cutout_basic(
                #data_np, 0, 0, wd - 1, ht - 1, nscale_x, nscale_y,
                data_np, 0, 0, wd, ht, nscale_x, nscale_y,
                logger=self.logger)

        mydata = self.baseimage._get_data()
        method = self.t_['mosaic_method']

        if method == 'simple':
            self.logger.debug("plotting by rotating/flipping image by WCS")
            # CASE 1: simple rotation and flips
            # Rotate piece into our orientation, according to wcs
            rot_dx, rot_dy = xrot - self.xrot_ref, yrot - self.yrot_ref

            flip_x = False
            flip_y = False

            # Optomization for 180 rotations
            if (np.isclose(math.fabs(rot_dx), 180.0) or
                np.isclose(math.fabs(rot_dy), 180.0)):
                rotdata = trcalc.transform(data_np,
                                           flip_x=True, flip_y=True)
                rot_dx = 0.0
                rot_dy = 0.0
            else:
                rotdata = data_np

            # convert to same type as basedata
            rotdata = rotdata.astype(mydata.dtype)

            # add an alpha layer
            minv, maxv = trcalc.get_minmax_dtype(rotdata.dtype)
            rotdata = trcalc.add_alpha(rotdata, alpha=maxv)

            # Finish with any necessary rotation of piece
            if not np.isclose(rot_dy, 0.0):
                rot_deg = rot_dy
                self.logger.debug("rotating %s by %f deg" % (name, rot_deg))
                rotdata = trcalc.rotate(rotdata, rot_deg,
                                        #rotctr_x=ctr_x, rotctr_y=ctr_y,
                                        logger=self.logger, pad=0)

            # Flip X due to negative CDELT1
            if np.sign(cdelt1) != np.sign(self.cdelt1_ref):
                flip_x = True

            # Flip Y due to negative CDELT2
            if np.sign(cdelt2) != np.sign(self.cdelt2_ref):
                flip_y = True

            if flip_x or flip_y:
                rotdata = trcalc.transform(rotdata,
                                           flip_x=flip_x, flip_y=flip_y)

            # Get size and data of new image
            ht, wd = rotdata.shape[:2]
            ctr_x, ctr_y = trcalc.get_center(rotdata)

            # Find location of image piece (center) in our array
            x0, y0 = self.baseimage.radectopix(ra, dec, coords='data')

            # Merge piece as closely as possible into our array
            # Unfortunately we lose a little precision rounding to the
            # nearest pixel--can't be helped with this approach
            x0, y0 = int(np.rint(x0)), int(np.rint(y0))
            self.logger.debug("Fitting image '%s' into mosaic at %f,%f" % (
                name, x0, y0))

            # This is for useful debugging info only
            my_ctr_x, my_ctr_y = trcalc.get_center(mydata)
            off_x, off_y = x0 - my_ctr_x, y0 - my_ctr_y
            self.logger.debug("centering offsets: %d,%d" % (off_x, off_y))

            # Sanity check piece placement
            xlo, xhi = x0 - ctr_x, x0 + wd - ctr_x
            ylo, yhi = y0 - ctr_y, y0 + ht - ctr_y
            assert (xhi - xlo == wd), \
                Exception("Width differential %d != %d" % (xhi - xlo, wd))
            assert (yhi - ylo == ht), \
                Exception("Height differential %d != %d" % (yhi - ylo, ht))

        elif method == 'warp':
            # convert to same type as basedata
            data_np = data_np.astype(mydata.dtype)

            self.logger.debug("plotting by warping image according to WCS")
            # CASE 2: user wants precise transformation of image using WCS
            dst, old_pts, coords, new_pts, dst_pts = warp_image(data_np,
                                                                image.wcs,
                                                                self.baseimage.wcs)

            # Merge piece as closely as possible into our array
            # Unfortunately we lose a little precision rounding to the
            # nearest pixel--can't be helped with this approach
            xlo, ylo = np.rint(new_pts[0] - dst_pts[0]).astype(int)
            self.logger.debug("Fitting image '%s' into mosaic at %f,%f" % (
                name, xlo, ylo))

            ht, wd = dst.shape[:2]
            xhi, yhi = xlo + wd, ylo + ht

            rotdata = dst

        else:
            raise ValueError(f"don't understand mosaic method '{method}'")

        #-----------
        mywd, myht = self.baseimage.get_size()
        if xlo < 0 or xhi > mywd or ylo < 0 or yhi > myht:
            if not self.t_['allow_expand']:
                raise Exception("New piece doesn't fit on image and "
                                "allow_expand=False")

            # <-- Resize our data array to allow the new image

            # determine amount to pad expansion by
            expand_x = max(int(self.t_['expand_pad_deg'] / self.scale_x), 0)
            expand_y = max(int(self.t_['expand_pad_deg'] / self.scale_y), 0)

            nx1_off, nx2_off = 0, 0
            if xlo < 0:
                nx1_off = abs(xlo) + expand_x
            if xhi > mywd:
                nx2_off = (xhi - mywd) + expand_x
            xlo, xhi = xlo + nx1_off, xhi + nx1_off

            ny1_off, ny2_off = 0, 0
            if ylo < 0:
                ny1_off = abs(ylo) + expand_y
            if yhi > myht:
                ny2_off = (yhi - myht) + expand_y
            ylo, yhi = ylo + ny1_off, yhi + ny1_off

            new_wd = mywd + nx1_off + nx2_off
            new_ht = myht + ny1_off + ny2_off

            # sanity check on new mosaic size
            old_area = mywd * myht
            new_area = new_wd * new_ht
            expand_pct = new_area / old_area
            if ((self.t_['max_expand_pct'] is not None) and
                    (expand_pct > self.t_['max_expand_pct'])):
                raise Exception("New area exceeds current one by %.2f %%;"
                                "increase max_expand_pct (%.2f) to allow" %
                                (expand_pct * 100, self.t_['max_expand_pct']))

            # go for it!
            #new_data = np.zeros((new_ht, new_wd))
            new_data = np.full((new_ht, new_wd, 2), np.nan, dtype=mydata.dtype)
            new_data[..., 1] = 0.0

            # place current data into new data
            new_data[ny1_off:ny1_off + myht, nx1_off:nx1_off + mywd] = mydata
            self.baseimage._data = new_data
            mydata = new_data

            if (nx1_off > 0) or (ny1_off > 0):
                # Adjust our WCS for relocation of the reference pixel
                crpix1, crpix2 = self.baseimage.get_keywords_list('CRPIX1', 'CRPIX2')
                kwds = dict(CRPIX1=crpix1 + nx1_off,
                            CRPIX2=crpix2 + ny1_off,
                            NAXIS1=new_wd, NAXIS2=new_ht)
                self.baseimage.update_keywords(kwds)

        # fit image piece into our array
        try:
            if self.t_['merge']:
                mydata[ylo:yhi, xlo:xhi, ...] += rotdata[0:ht, 0:wd, ...]
            else:
                mask = (mydata[ylo:yhi, xlo:xhi, 1] <= 0.0)
                mydata[ylo:yhi, xlo:xhi, ...][mask] = rotdata[0:ht, 0:wd, ...][mask]

        except Exception as e:
            self.logger.error("Error fitting tile: %s" % (str(e)))
            raise

        return (xlo, ylo, xhi, yhi)

    def ingest_one(self, image):
        """Ingest an image in the right place in the image array.

        This method is typically called internally.
        """
        llur = self.ingest_image(image)

        self.make_callback('progress', 'fitting',
                           float(self.ingest_count) / self.total_images)

    def reset(self):
        """Prepare for a new mosaic.
        The next call to ```mosaic`` will create a new mosaic.
        """
        self.baseimage = None
        self.image_list = []
        self.ingest_count = 0
        self.total_images = 0

    def annotate_images(self, canvas):
        tagpfx = self.t_['ann_tag_pfx']
        tags = canvas.get_tags_by_tag_pfx(tagpfx)
        canvas.delete_objects_by_tag(tags, redraw=False)

        if self.t_['annotate_images']:
            dc = canvas.get_draw_classes()
            for name, tag, ra, dec in self.image_list:
                x, y = self.baseimage.radectopix(ra, dec)
                text = dc.Text(x, y, name,
                               color=self.t_['annotate_color'],
                               fontsize=self.t_['annotate_fontsize'],
                               fontscale=True)
                tag = tagpfx + tag
                canvas.add(text, tag=tag, redraw=False)

        canvas.update_canvas(whence=3)

    def mosaic(self, images, ev_intr=None):
        """Create a mosaic of ``images``.

        Returns a `~ginga.AstroImage.AstroImage`
        """
        num_images = len(images)
        if num_images == 0:
            return
        self.total_images += num_images

        self.make_callback('progress', 'fitting', 0.0)
        t1 = time.time()

        # If there is no current mosaic then prepare a new one
        if self.baseimage is None:
            ref_image = images[0]
            self.prepare_mosaic(ref_image)

        self.logger.info("fitting tiles...")

        for image in images:
            if ev_intr is not None and ev_intr.is_set():
                raise Exception("interrupted by user")
            self.ingest_one(image)

        self.logger.info("finishing...")
        self.make_callback('progress', 'finishing', 0.0)

        self.process_elapsed = time.time() - t1
        self.logger.info("mosaic done. process=%.4f (sec)" % (
            self.process_elapsed))

        self.make_callback('finished', self.process_elapsed)

        return self.baseimage


class CanvasMosaicer(Callback.Callbacks):
    """Class for creating collages on a Ginga canvas.

    A collage is sort of like a mosaic, except that instead of creating a
    large image array, individual tiles are transformed and plotted on a
    canvas.

    Typical usage:

    >>> collager = CanvasMosaicer(logger)
    >>> collager.mosaic(viewer, images)

    where ``images`` is a list of `~ginga.AstroImage.AstroImage` that
    should be plotted in ``viewer``.
    """
    def __init__(self, logger, settings=None):
        super(CanvasMosaicer, self).__init__()

        self.logger = logger

        if settings is None:
            settings = Settings.SettingGroup(name='collager',
                                             logger=self.logger)
        # options
        self.t_ = settings
        self.t_.set_defaults(annotate_images=False, annotate_color='pink',
                             annotate_fontsize=10.0, ann_fits_kwd=None,
                             ann_tag_pfx='ann_',
                             match_bg=False, collage_method='simple',
                             center_image=False)

        self.ingest_count = 0
        # holds processed images to be inserted into mosaic image
        self.total_images = 0

        # these are updated in prepare_mosaic() and represent measurements
        # on the reference image
        self.bg_ref = 0.0
        self.xrot_ref, self.yrot_ref = 0.0, 0.0
        self.cdelt1_ref, self.cdelt2_ref = 1.0, 1.0
        self.scale_x = 1.0
        self.scale_y = 1.0
        self.limits = None
        self.ref_image = None
        self.image_list = []

        for name in ['progress', 'finished']:
            self.enable_callback(name)

    def get_settings(self):
        return self.t_

    def prepare_mosaic(self, ref_image):
        """Prepare for a new mosaic image based on the pointing of
        the reference image ``ref_image``.

        This method is typically called internally.
        """
        self.ref_image = ref_image

        # if user requesting us to match backgrounds, then calculate
        # median of root image and save it
        if self.t_['match_bg']:
            data_np = ref_image.get_data()
            self.bg_ref = iqcalc.get_median(data_np)

        header = ref_image.get_header()

        # TODO: handle skew (differing rotation for each axis)?
        (rot_xy, cdelt_xy) = wcs.get_xy_rotation_and_scale(header)
        self.logger.debug("ref image rot_x=%f rot_y=%f cdelt1=%f cdelt2=%f" % (
            rot_xy[0], rot_xy[1], cdelt_xy[0], cdelt_xy[1]))

        # Store base image rotation and scale
        self.xrot_ref, self.yrot_ref = rot_xy
        self.cdelt1_ref, self.cdelt2_ref = cdelt_xy
        self.scale_x = math.fabs(cdelt_xy[0])
        self.scale_y = math.fabs(cdelt_xy[1])
        self.limits = ((0, 0), (0, 0))

    def _get_name_tag(self, image):
        self.ingest_count += 1
        tag = 'image{}'.format(self.ingest_count)

        ann_fits_kwd = self.t_['ann_fits_kwd']
        if ann_fits_kwd is not None:
            header = image.get_header()
            name = str(header[ann_fits_kwd])
        else:
            name = image.get('name', tag)

        tag = image.get('tag', tag)
        return (name, tag)

    def transform_image(self, image):
        """Prepare ``image`` to be plotted in the right place according to
        the reference image WCS.  A new image is returned.

        This method is typically called internally.
        """
        name, tag = self._get_name_tag(image)

        data_np = image._get_data()
        if 0 in data_np.shape:
            self.logger.info("Skipping image with zero length axis")
            return

        ht, wd = data_np.shape

        # If caller asked us to match background of pieces then
        # fix up this data
        if self.t_['match_bg']:
            bg = iqcalc.get_median(data_np)
            bg_inc = self.bg_ref - bg
            data_np = data_np + bg_inc

        # Calculate sky position at the center of the piece
        ctr_x, ctr_y = trcalc.get_center(data_np)
        ra, dec = image.pixtoradec(ctr_x, ctr_y, coords='data')
        self.image_list.append((name, tag, ra, dec))

        # Get rotation and scale of piece
        header = image.get_header()
        ((xrot, yrot),
         (cdelt1, cdelt2)) = wcs.get_xy_rotation_and_scale(header)
        self.logger.debug("image(%s) xrot=%f yrot=%f cdelt1=%f "
                          "cdelt2=%f" % (name, xrot, yrot, cdelt1, cdelt2))

        # scale if necessary to scale of reference image
        if (not np.isclose(math.fabs(cdelt1), self.scale_x) or
            not np.isclose(math.fabs(cdelt2), self.scale_y)):
            nscale_x = math.fabs(cdelt1) / self.scale_x
            nscale_y = math.fabs(cdelt2) / self.scale_y
            self.logger.debug("scaling piece by x(%f), y(%f)" % (
                nscale_x, nscale_y))
            data_np, (ascale_x, ascale_y) = trcalc.get_scaled_cutout_basic(
                #data_np, 0, 0, wd - 1, ht - 1, nscale_x, nscale_y,
                data_np, 0, 0, wd, ht, nscale_x, nscale_y,
                logger=self.logger)

        method = self.t_['collage_method']

        if method == 'simple':
            self.logger.debug("plotting by rotating/flipping image by WCS")
            # CASE 1: simple rotation and flips
            # Rotate piece into our orientation, according to wcs
            rot_dx, rot_dy = xrot - self.xrot_ref, yrot - self.yrot_ref

            flip_x = False
            flip_y = False

            # Optomization for 180 rotations
            if (np.isclose(math.fabs(rot_dx), 180.0) or
                np.isclose(math.fabs(rot_dy), 180.0)):
                rotdata = trcalc.transform(data_np,
                                           flip_x=True, flip_y=True)
                rot_dx = 0.0
                rot_dy = 0.0
            else:
                rotdata = data_np

            # Finish with any necessary rotation of piece
            ignore_alpha = False
            if not np.isclose(rot_dy, 0.0):
                rot_deg = rot_dy
                minv, maxv = trcalc.get_minmax_dtype(rotdata.dtype)
                rotdata = trcalc.add_alpha(rotdata, alpha=maxv)
                self.logger.debug("rotating %s by %f deg" % (name, rot_deg))
                rotdata = trcalc.rotate(rotdata, rot_deg,
                                        #rotctr_x=ctr_x, rotctr_y=ctr_y,
                                        logger=self.logger, pad=0)
                ignore_alpha = True

            # Flip X due to negative CDELT1
            if np.sign(cdelt1) != np.sign(self.cdelt1_ref):
                flip_x = True

            # Flip Y due to negative CDELT2
            if np.sign(cdelt2) != np.sign(self.cdelt2_ref):
                flip_y = True

            if flip_x or flip_y:
                rotdata = trcalc.transform(rotdata,
                                           flip_x=flip_x, flip_y=flip_y)

            # new wrapper for transformed image
            metadata = dict(header=header, ignore_alpha=ignore_alpha)
            new_image = AstroImage.AstroImage(data_np=rotdata, metadata=metadata)

            # Get size and data of new image
            ht, wd = rotdata.shape[:2]
            ctr_x, ctr_y = trcalc.get_center(rotdata)

            # Find location of image piece (center) in our array
            x0, y0 = self.ref_image.radectopix(ra, dec, coords='data')

            #x0, y0 = int(np.round(x0)), int(np.round(y0))
            self.logger.debug("Fitting image '%s' into mosaic at %f,%f" % (
                name, x0, y0))

            # update limits
            xlo, xhi = x0 - ctr_x, x0 + ctr_x
            ylo, yhi = y0 - ctr_y, y0 + ctr_y

        elif method == 'warp':
            # Need to convert to float for this type
            data_np = data_np.astype(np.float32)

            self.logger.debug("plotting by warping image according to WCS")
            # CASE 2: user wants precise transformation of image using WCS
            dst, old_pts, coords, new_pts, dst_pts = warp_image(data_np,
                                                                image.wcs,
                                                                self.ref_image.wcs)

            # new wrapper for transformed image
            metadata = dict(header=header, ignore_alpha=True)
            new_image = AstroImage.AstroImage(data_np=dst, metadata=metadata)

            # find x, y at which to plot image
            xlo, ylo = new_pts[0] - dst_pts[0]
            self.logger.debug("Fitting image '%s' into mosaic at %f,%f" % (
                name, xlo, ylo))

            new_ht, new_wd = dst.shape[:2]
            xhi, yhi = xlo + new_wd, ylo + new_ht

        else:
            raise ValueError(f"don't understand mosaic method '{method}'")

        new_image.set(xpos=xlo, ypos=ylo, name=name, tag=tag)

        # calculate new limits of canvas
        _xlo, _ylo, = self.limits[0]
        _xhi, _yhi, = self.limits[1]
        _xlo, _ylo = min(_xlo, xlo), min(_ylo, ylo)
        _xhi, _yhi = max(_xhi, xhi), max(_yhi, yhi)
        self.limits = ((_xlo, _ylo), (_xhi, _yhi))

        return new_image

    def annotate_images(self, canvas):
        tagpfx = self.t_['ann_tag_pfx']
        tags = canvas.get_tags_by_tag_pfx(tagpfx)
        canvas.delete_objects_by_tag(tags, redraw=False)

        if self.t_['annotate_images']:
            dc = canvas.get_draw_classes()
            for name, tag, ra, dec in self.image_list:
                x, y = self.ref_image.radectopix(ra, dec)
                text = dc.Text(x, y, name,
                               color=self.t_['annotate_color'],
                               fontsize=self.t_['annotate_fontsize'],
                               fontscale=True)
                tag = tagpfx + tag
                canvas.add(text, tag=tag, redraw=False)

        canvas.update_canvas(whence=3)

    def plot_image(self, canvas, image):
        """Plot a new image created by ``transform_image()`` on ``canvas``.

        This is typically called internally.
        """
        dc = canvas.get_draw_classes()
        xpos, ypos, name, tag = image.get_list('xpos', 'ypos',
                                               'name', 'tag')
        img = dc.NormImage(xpos, ypos, image)
        img.is_data = True
        canvas.add(img, tag=tag, redraw=False)

    def reset(self):
        """Prepare for a new mosaic.
        The next call to ```mosaic`` will create a new mosaic.
        """
        self.ref_image = None
        self.ingest_count = 0
        self.total_images = 0
        self.image_list = []

    def ingest_one(self, canvas, image):
        """Plot ``image`` in the right place on the ``canvas``.

        This is typically called internally.
        """
        new_image = self.transform_image(image)

        self.plot_image(canvas, new_image)

        self.make_callback('progress', 'fitting',
                           float(self.ingest_count) / self.total_images)

    def mosaic(self, viewer, images, canvas=None, ev_intr=None):
        """Plot a mosaic of ``images`` in ``viewer`` on ``canvas``.
        If ``canvas`` is `None` the viewer's default canvas is used.
        """
        images = list(images)    # because we might pop(0)
        num_images = len(images)
        if num_images == 0:
            return
        self.total_images += num_images

        self.make_callback('progress', 'fitting', 0.0)
        t1 = time.time()

        if canvas is None:
            canvas = viewer.get_canvas()

        with viewer.suppress_redraw:
            # If there is no current mosaic then prepare a new one
            if self.ref_image is None:
                ref_image = images.pop(0)
                name, tag = self._get_name_tag(ref_image)
                self.prepare_mosaic(ref_image)

                # TODO: delete only items we may have added
                canvas.delete_all_objects(redraw=False)
                # first image is loaded in the usual way
                viewer.set_image(ref_image)
                self.limits = viewer.get_limits()

                # save position of reference image for annotation
                name, tag = self._get_name_tag(ref_image)
                wd, ht = ref_image.get_size()
                ctr_x, ctr_y = wd * 0.5, ht * 0.5
                ctr_ra, ctr_dec = ref_image.pixtoradec(ctr_x, ctr_y)
                self.image_list.append((name, tag, ctr_ra, ctr_dec))

            self.logger.info("fitting tiles...")

            for image in images:
                time.sleep(0)
                if ev_intr is not None and ev_intr.is_set():
                    raise Exception("interrupted by user")
                self.ingest_one(canvas, image)

            self.logger.info("finishing...")
            self.annotate_images(canvas)
            self.make_callback('progress', 'finishing', 0.0)

            viewer.set_limits(self.limits)
            if self.t_['center_image']:
                viewer.center_image()
            canvas.update_canvas(whence=0)

        self.process_elapsed = time.time() - t1
        self.logger.info("collage done. process=%.4f (sec)" % (
            self.process_elapsed))

        self.make_callback('finished', self.process_elapsed)



# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
"""
These pipeline stages are used in creating the StandardPipelineRenderer,
the default renderer for standard widget backends.

[createbg] => [overlays] => [iccprof] => [flipswap] => [rotate] => [output]

"""
import numpy as np

from ginga import trcalc, RGBImage

from .base import Stage, StageError


class CreateBg(Stage):
    """Create the background RGB image, sized to fit the area that needs to
    be painted in the viewer, and big enough so there is room to rotate it.
    """

    _stagename = 'viewer-createbg'

    def __init__(self, viewer):
        super().__init__()

        self.viewer = viewer
        self.dtype = np.uint8

    def run(self, prev_stage):
        if prev_stage is not None:
            raise StageError("'{}' in wrong location".format(self._stagename))

        if self._bypass:
            self.pipeline.send(res_np=None)
            return

        state = self.pipeline.get('state')
        win_wd, win_ht = state.win_dim

        # calc minimum size of pixel image we will generate
        # necessary to fit the window in the desired size

        # Make a square from the scaled cutout, with room to rotate
        slop = 20
        side = int(np.sqrt(win_wd**2 + win_ht**2) + slop)
        wd = ht = side

        # Find center of new array
        ncx, ncy = wd // 2, ht // 2
        depth = len(state.order)

        # make backing image with the background color
        r, g, b = self.viewer.get_bg()
        res_np = trcalc.make_filled_array((ht, wd, depth), self.dtype,
                                          state.order, r, g, b, 1.0)

        self.pipeline.set(org_dim=(wd, ht), org_off=(ncx, ncy))
        self.pipeline.send(res_np=res_np)


class ICCProf(Stage):
    """Convert the given RGB data from the input ICC profile to the
    output ICC profile registered for the viewer.
    """
    _stagename = 'viewer-icc-profiler'

    def __init__(self, viewer):
        super().__init__()

        self.viewer = viewer

    def run(self, prev_stage):
        data = self.pipeline.get_data(prev_stage)
        self.verify_2d(data)

        from ginga.util import rgb_cms

        working_profile = rgb_cms.working_profile
        t_ = self.viewer.get_settings()
        output_profile = t_.get('icc_output_profile', None)

        if (self._bypass or not rgb_cms.have_cms or
            None in [working_profile, output_profile]):
            self.pipeline.set(icc_output_profile=working_profile)
            self.pipeline.send(res_np=data)
            return

        # get rest of necessary conversion parameters
        to_intent = t_.get('icc_output_intent', 'perceptual')
        proofprof_name = t_.get('icc_proof_profile', None)
        proof_intent = t_.get('icc_proof_intent', 'perceptual')
        use_black_pt = t_.get('icc_black_point_compensation', False)
        try:
            data = rgb_cms.convert_profile_fromto(data,
                                                  working_profile,
                                                  output_profile,
                                                  to_intent=to_intent,
                                                  proof_name=proofprof_name,
                                                  proof_intent=proof_intent,
                                                  use_black_pt=use_black_pt,
                                                  logger=self.logger)

            self.logger.debug("Converted from '%s' to '%s' profile" % (
                working_profile, output_profile))

        except Exception as e:
            self.logger.warning("Error converting output from working profile: %s" % (str(e)))
            # TODO: maybe should have a traceback here
            self.logger.info("Output left unprofiled")

        self.pipeline.set(icc_output_profile=output_profile)
        self.pipeline.send(res_np=data)


class FlipSwap(Stage):
    """Flip or swap axes according the viewer settings.
    """

    _stagename = 'viewer-flipswap'

    def __init__(self, viewer):
        super().__init__()

        self.viewer = viewer

    def run(self, prev_stage):
        data = self.pipeline.get_data(prev_stage)
        self.verify_2d(data)

        xoff, yoff = self.pipeline.get('org_off')
        if not self._bypass:
            flip_x, flip_y, swap_xy = self.viewer.get_transforms()

            ht, wd = data.shape[:2]

            # Do transforms as necessary
            data = trcalc.transform(data, flip_x=flip_x, flip_y=flip_y,
                                    swap_xy=swap_xy)
            if flip_y:
                yoff = ht - yoff
            if flip_x:
                xoff = wd - xoff
            if swap_xy:
                xoff, yoff = yoff, xoff

        self.pipeline.set(off=(xoff, yoff))
        self.pipeline.send(res_np=data)


class Rotate(Stage):
    """Rotate the image according the viewer settings.
    """

    _stagename = 'viewer-rotate'

    def __init__(self, viewer):
        super().__init__()

        self.viewer = viewer

    def run(self, prev_stage):
        data = self.pipeline.get_data(prev_stage)
        self.verify_2d(data)

        if not self._bypass:
            rot_deg = self.viewer.get_rotation()

            if not np.isclose(rot_deg, 0.0):
                data = np.copy(data)
                #data = np.ascontiguousarray(data)
                data = trcalc.rotate_clip(data, -rot_deg, out=data,
                                          logger=self.logger)

            # apply other transforms
            if self.viewer._invert_y:
                # Flip Y for natural Y-axis inversion between FITS coords
                # and screen coords
                data = np.flipud(data)

        # dimensions may have changed in transformations
        ht, wd = data.shape[:2]
        xoff, yoff = self.pipeline.get('off')
        state = self.pipeline.get('state')
        ctr_x, ctr_y = state.ctr

        dst_x, dst_y = ctr_x - xoff, ctr_y - (ht - yoff)
        self.pipeline.set(dst=(dst_x, dst_y))
        self.logger.debug("ctr=%d,%d off=%d,%d dst=%d,%d cutout=%dx%d" % (
            ctr_x, ctr_y, xoff, yoff, dst_x, dst_y, wd, ht))

        win_wd, win_ht = state.win_dim
        self.logger.debug("win=%d,%d coverage=%d,%d" % (
            win_wd, win_ht, dst_x + wd, dst_y + ht))

        self.pipeline.send(res_np=data)


class Output(Stage):
    """Create the output image by trimming height and width and reordering
    channels according to viewer settings.
    """

    _stagename = 'viewer-output'

    def __init__(self, viewer):
        super().__init__()

        self.viewer = viewer

    def run(self, prev_stage):
        data = self.pipeline.get_data(prev_stage)

        ## assert (len(data.shape) == 3
        ##         and data.dtype == np.dtype(np.uint8)
        ##         and data.shape[2] in [3, 4]), \
        ##     StageError("Expecting a RGB[A] image in final stage")
        self.verify_2d(data)

        state = self.pipeline.get('state')
        out_order = state.order

        if not self._bypass:
            ht, wd = data.shape[:2]
            state = self.pipeline.get('state')
            win_wd, win_ht = state.win_dim
            if wd < win_wd or ht < win_ht:
                raise StageError("pipeline output doesn't cover window")

            # now cut out the size that we need
            dst_x, dst_y = self.pipeline.get('dst')
            if dst_x > 0 or dst_y > 0:
                raise StageError("pipeline calculated dst is not correct")

            x1, y1 = abs(dst_x), abs(dst_y)
            data = data[y1:y1 + win_ht, x1:x1 + win_wd]

            # reorder image for renderer's desired format
            dst_order = self.viewer.renderer.get_rgb_order()
            data = trcalc.reorder_image(dst_order, data, state.order)
            data = np.ascontiguousarray(data)
            out_order = dst_order

        self.pipeline.set(out_order=out_order)
        self.pipeline.send(res_np=data)


class Overlays(Stage):
    """The Overlays stage constructs an embedded pipeline to process each
    canvas into a RGBA image.  A pipeline is run for each image that
    needs to be processed. The pipeline run depends on the type of image
    on the canvas:

    ginga.canvas.types.image.Image:

    [scale] => [reorder] => [merge]

    ginga.canvas.types.image.NormImage:

    [scale] => [reorder] => [cuts] => [rgbmap] => [merge]
    """

    _stagename = 'viewer-image-overlays'

    def __init__(self, viewer):
        super(Overlays, self).__init__()

        self.viewer = viewer

    def run(self, prev_stage):
        whence = self.pipeline.get('whence')
        dstarr = self.pipeline.get_data(self)
        if whence >= 3 and dstarr is not None:
            # if we are only overplotting (whence >= 3) then we can
            # simply use the result of the last pipeline run because
            # nothing should have changed
            self.pipeline.send(res_np=dstarr)
            return

        bgarr = self.pipeline.get_data(prev_stage)
        self.verify_2d(bgarr)

        dstarr = np.copy(bgarr)
        self.pipeline.set(dstarr=dstarr)

        p_canvas = self.viewer.get_private_canvas()
        self._overlay_images(p_canvas, whence=whence)

        self.pipeline.send(res_np=dstarr)

    def _overlay_images(self, canvas, whence=0.0):
        if not hasattr(canvas, 'objects'):
            return

        for obj in canvas.get_objects():
            if hasattr(obj, 'prepare_image'):
                obj.prepare_image(self.viewer, whence)
            elif obj.is_compound() and (obj != canvas):
                self._overlay_images(obj, whence=whence)

    def _prepare_image(self, cvs_img, cache, whence):
        from ginga.util import pipeline
        pipe = cache.get('minipipe', None)
        if pipe is None:
            stages = [Scale(self.viewer),    # 0
                      Reorder(self.viewer),  # 1
                      Merge(self.viewer)]    # 2
            pipe = pipeline.Pipeline(self.logger, stages)
            pipe.name = 'image-overlays'
            cache.minipipe = pipe
        state = self.pipeline.get('state')
        pipe.set(whence=whence, cvs_img=cvs_img, state=state,
                 dstarr=self.pipeline.get('dstarr'))
        if whence <= 0:
            pipe.run_from(pipe[0])
            return

        if not cache.visible:
            # set by Scale stage depending on image visibility in result
            return
        pipe.run_from(pipe[1])

    def _prepare_norm_image(self, cvs_img, cache, whence):
        from ginga.util import pipeline
        pipe = cache.get('minipipe', None)
        if pipe is None:
            stages = [Scale(self.viewer),    # 0
                      Reorder(self.viewer),  # 1
                      Cuts(self.viewer),     # 2
                      RGBMap(self.viewer),   # 3
                      Merge(self.viewer)]    # 4
            pipe = pipeline.Pipeline(self.logger, stages)
            pipe.name = 'image-overlays'
            cache.minipipe = pipe
        state = self.pipeline.get('state')
        pipe.set(whence=whence, cvs_img=cvs_img, state=state,
                 dstarr=self.pipeline.get('dstarr'))
        if whence <= 0:
            pipe.run_from(pipe[0])
            return

        if not cache.visible:
            # set by Scale stage depending on image visibility in result
            return
        elif whence <= 1:
            pipe.run_from(pipe[2])  # Cuts
        elif whence <= 2:
            pipe.run_from(pipe[3])  # RGBMap
        else:
            pipe.run_from(pipe[4])  # Merge


class Scale(Stage):
    """Scale the image according to the viewers scale settings.  Only the
    portions of the image that will be visible are copied and then scaled.
    """

    _stagename = 'viewer-scale'

    def __init__(self, viewer):
        super().__init__()

        self.viewer = viewer

    def run(self, prev_stage):
        #if prev_stage is not None:
        #    raise StageError(f"'{}' in wrong location".format(self._stagename))
        cvs_img = self.pipeline.get('cvs_img')
        cache = cvs_img.get_cache(self.viewer)

        image = cvs_img.get_image()
        if image is None:
            self.pipeline.send(res_np=None)
            return

        data_np = image.get_data()
        self.verify_2d(data_np)

        # get extent of our data coverage in the window
        # TODO: get rid of padding by fixing get_draw_bbox() which
        # doesn't quite get the coverage right at high magnifications
        pad = 1.0
        pts = np.asarray(self.viewer.get_draw_bbox()).T
        xmin = np.floor(np.nanmin(pts[0])) - pad
        ymin = np.floor(np.nanmin(pts[1])) - pad
        xmax = np.ceil(np.nanmax(pts[0])) + pad
        ymax = np.ceil(np.nanmax(pts[1])) + pad

        # NOTE: if we see NaNs here, consider the pipeline to be terminated
        # with nothing to draw
        if True in np.isnan([xmin, ymin, xmax, ymax]):
            # no overlay needed
            self.pipeline.send(res_np=None)
            cache.visible = False
            self.pipeline.stop()
            return

        xmin, ymin, xmax, ymax = int(xmin), int(ymin), int(xmax), int(ymax)

        # get destination location in data_coords
        img = cvs_img
        dst_x, dst_y = img.crdmap.to_data((img.x, img.y))

        ht, wd = data_np.shape[:2]
        # TODO: think we need to apply scaling factors to wd/ht
        # BEFORE we calculate merge clip
        a1, b1, a2, b2 = 0, 0, wd - 1, ht - 1

        # calculate the cutout that we can make and scale to merge
        # onto the final image--by only cutting out what is necessary
        # this speeds scaling greatly at zoomed in sizes
        ((dst_x, dst_y), (a1, b1), (a2, b2)) = \
            trcalc.calc_image_merge_clip((xmin, ymin), (xmax, ymax),
                                         (dst_x, dst_y),
                                         (a1, b1), (a2, b2))

        # is image completely off the screen?
        if (a2 - a1 <= 0) or (b2 - b1 <= 0):
            # no overlay needed
            self.pipeline.send(res_np=None)
            cache.visible = False
            self.pipeline.stop()
            return

        cache.visible = True

        # cutout and scale the piece appropriately by the viewer scale
        scale_x, scale_y = self.viewer.get_scale_xy()
        # scale additionally by scale specified in canvas image
        _scale_x, _scale_y = (scale_x * img.scale_x,
                              scale_y * img.scale_y)

        interp = img.interpolation
        if interp is None:
            t_ = self.viewer.get_settings()
            interp = t_.get('interpolation', 'basic')
        if interp not in trcalc.interpolation_methods:
            interp = 'basic'

        data, scales = trcalc.get_scaled_cutout_basic(data_np, a1, b1, a2, b2,
                                                      _scale_x, _scale_y,
                                                      interpolation=interp,
                                                      logger=self.logger)

        if img.flipy:
            data = np.flipud(data)

        # calculate our offset from the pan position
        pan_x, pan_y = self.viewer.get_pan()
        pan_off = self.viewer.data_off
        pan_x, pan_y = pan_x + pan_off, pan_y + pan_off
        off_x, off_y = dst_x - pan_x, dst_y - pan_y
        # scale offset
        off_x *= scale_x
        off_y *= scale_y

        self.pipeline.set(offset=(off_x, off_y))

        self.pipeline.send(res_np=data)


class Reorder(Stage):
    """Reorder RGB images for the pipeline.

    Reorders multiband images for processing in the pipeline standard
    order.  Additionally, if the image has an alpha channel, then strip
    it off and save it until it is recombined later with the colorized
    output--this saves us having to deal with an alpha band in the cuts
    leveling and RGB mapping routines.
    """

    _stagename = 'viewer-reorder'

    def __init__(self, viewer):
        super(Reorder, self).__init__()

        self.viewer = viewer

    def run(self, prev_stage):
        data = self.pipeline.get_data(prev_stage)
        if data is None:
            self.pipeline.send(res_np=None)
            return
        self.verify_2d(data)

        cvs_img = self.pipeline.get('cvs_img')

        state = self.pipeline.get('state')
        image = cvs_img.get_image()
        image_order = image.get_order()

        if isinstance(image, RGBImage.RGBImage) and image_order != state.order:
            # reorder image channels for pipeline
            data = trcalc.reorder_image(state.order, data, image_order)

        # strip off alpha layer, if any
        if 'A' not in image_order:
            alpha = None

        else:
            # strip off alpha layer
            # data, alpha = trcalc.remove_alpha(data)
            a_idx = image_order.index('A')
            alpha = data[..., a_idx]
            data = data[..., 0:a_idx]
            ht, wd, dp = data.shape
            if dp == 1:
                data = data.reshape((ht, wd))

        # currently added back on in Merge stage
        self.pipeline.set(alpha=alpha)

        self.pipeline.send(res_np=data)


class Merge(Stage):
    """Merge the processed image into the result of our stage at the
    appropriate offset and with the appropriate alpha blending.
    """

    _stagename = 'viewer-merge-overlay'

    def __init__(self, viewer):
        super().__init__()

        self.viewer = viewer

    def run(self, prev_stage):
        rgbarr = self.pipeline.get_data(prev_stage)
        if rgbarr is None:
            # nothing to merge
            return
        self.verify_2d(rgbarr)

        if len(rgbarr.shape) != 3 or rgbarr.shape[2] < 3:
            raise StageError("Not an RGB array (shape={})".format(rgbarr.shape))

        cvs_img = self.pipeline.get('cvs_img')
        off_x, off_y = self.pipeline.get('offset')
        dstarr = self.pipeline.get('dstarr')
        state = self.pipeline.get('state')

        # dst position in the pre-transformed array should be calculated
        # from the center of the array plus offsets
        ht, wd, dp = dstarr.shape
        cvs_x = int(np.round(wd * 0.5 + off_x))
        cvs_y = int(np.round(ht * 0.5 + off_y))
        cvs_pos = (cvs_x, cvs_y)

        dst_order = state.order
        image_order = state.order

        # merge back in alpha layer if one was stripped off earler
        # in the pipeline
        alpha = self.pipeline.get('alpha')
        if alpha is not None and 'A' in state.order:
            a_idx = state.order.index('A')
            if rgbarr.shape[2] != 4:
                #raise StageError("RGB array lacks alpha band (shape={})".format(rgbarr.shape))
                # insert an empty alpha layer to accomodate
                rgbarr = np.insert(rgbarr, a_idx,
                                   np.zeros(rgbarr.shape[:2],
                                            dtype=rgbarr.dtype),
                                   axis=2)
            # normalize alpha array to the final output range
            alpha = trcalc.array_convert(alpha, rgbarr.dtype)
            rgbarr[..., a_idx] = alpha

        # composite the image into the destination array at the
        # calculated position
        trcalc.overlay_image(dstarr, cvs_pos, rgbarr,
                             dst_order=dst_order, src_order=image_order,
                             # NOTE: these actually not used because rgbarr
                             # contains an alpha channel
                             alpha=cvs_img.alpha, fill=True,
                             flipy=False)   # cvs_img.flipy

        cache = cvs_img.get_cache(self.viewer)
        cache.drawn = True
        #self.pipeline.send(res_np=None)


class Cuts(Stage):
    """Process the cut levels (loval, hival) for the image.

    This has the effect of mapping every input value between loval and
    hival to the range specified by the output values (vmin, vmax).
    Input values lower than loval are mapped to loval and values higher
    than hival are mapped to hival.
    """

    _stagename = 'viewer-cut-levels'

    def __init__(self, viewer):
        super().__init__()

        self.viewer = viewer

    def run(self, prev_stage):
        data = self.pipeline.get_data(prev_stage)
        if data is None:
            self.pipeline.send(res_np=None)
            return
        self.verify_2d(data)

        cvs_img = self.pipeline.get('cvs_img')

        if cvs_img.rgbmap is not None:
            rgbmap = cvs_img.rgbmap
        else:
            rgbmap = self.viewer.get_rgbmap()

        vmin = 0
        vmax = rgbmap.get_hash_size() - 1

        if cvs_img.autocuts is not None:
            autocuts = cvs_img.autocuts
        else:
            autocuts = self.viewer.autocuts

        # Apply cut levels
        if cvs_img.cuts is not None:
            loval, hival = cvs_img.cuts
        else:
            loval, hival = self.viewer.t_['cuts']

        res_np = autocuts.cut_levels(data, loval, hival,
                                     vmin=vmin, vmax=vmax)

        # NOTE: optimization to prevent multiple coercions in
        # RGBMap
        if not np.issubdtype(res_np.dtype, np.uint):
            res_np = res_np.astype(np.uint)

        self.pipeline.send(res_np=res_np)


class RGBMap(Stage):
    """Color map an image using a RGBMap.

    This color-maps an image according to the settings in the RGBMap
    of the image (if it has one), or the viewer (if not).
    """

    _stagename = 'viewer-rgb-mapper'

    def __init__(self, viewer):
        super().__init__()

        self.viewer = viewer

    def run(self, prev_stage):
        arr_in = self.pipeline.get_data(prev_stage)
        if arr_in is None:
            self.pipeline.send(res_np=None)
            return
        self.verify_2d(arr_in)

        cvs_img = self.pipeline.get('cvs_img')
        state = self.pipeline.get('state')

        if cvs_img.rgbmap is not None:
            rgbmap = cvs_img.rgbmap
        else:
            rgbmap = self.viewer.get_rgbmap()

        # See NOTE in Cuts
        if not np.issubdtype(arr_in.dtype, np.dtype(np.uint)):
            arr_in = arr_in.astype(np.uint)

        # get RGB mapped array
        arr_out = rgbmap.get_rgb_array(arr_in, order=state.order)

        self.pipeline.send(res_np=arr_out)
